import json
import math
import tempfile
from enum import IntEnum
from pathlib import Path
from typing import Set

import matplotlib
import matplotlib.pyplot as plt
from flask import Markup, Response, make_response

from steady_cell_phenotype._graph_layout import graph_layout
from steady_cell_phenotype._util import *

matplotlib.use("agg")

Edge = Dict[str, Dict]


def get_trajectory_edge_list(
    *, init_state: np.ndarray, update_fn: Callable, target_variables: Tuple[str]
) -> Tuple[List[Edge], Set[HashableNdArray]]:
    trajectory: np.ndarray
    trajectory, limit_cycle = get_trajectory(init_state=init_state, update_fn=update_fn)

    edge_list = [
        {
            "source": dict(zip(target_variables, trajectory[i, :])),
            "target": dict(zip(target_variables, trajectory[i + 1, :])),
        }
        for i in range(trajectory.shape[0] - 1)
    ]
    if trajectory.shape[0] > 0:
        edge_list.append(
            {
                "source": dict(zip(target_variables, trajectory[-1, :])),
                "target": dict(zip(target_variables, limit_cycle[0, :])),
            }
        )
    edge_list.extend(
        {
            "source": dict(zip(target_variables, limit_cycle[i, :])),
            "target": dict(zip(target_variables, limit_cycle[i + 1, :])),
        }
        for i in range(limit_cycle.shape[0] - 1)
    )
    if limit_cycle.shape[0] > 0:
        edge_list.append(
            {
                "source": dict(zip(target_variables, limit_cycle[-1, :])),
                "target": dict(zip(target_variables, limit_cycle[0, :])),
            }
        )
        # TODO: can this be empty? I think not, unless I've forgotten what this is.
        #  Should throw an exception if ==0

    limit_cycle_set = set(
        HashableNdArray(limit_cycle[idx, :]) for idx in range(limit_cycle.shape[0])
    )

    return edge_list, limit_cycle_set


def add_nearby_states(init_states: List[np.ndarray]) -> List[np.ndarray]:
    """
    Adds all states Hamming distance 1 from an init state to the list
    Parameters
    ----------
    init_states

    Returns
    -------
    init_states plus hamming distance one states. Order not preserved.
    """
    if len(init_states) <= 0:
        return []

    new_init_states: Set[HashableNdArray] = set()
    num_vars: int = init_states[0].shape[0]

    # add all Hamming distance `dist` states from the initial states
    for state, idx, val in itertools.product(init_states, range(num_vars), range(3)):
        if abs(state[idx] - val) <= 1:
            new_state = state.copy()
            new_state[idx] = val
            new_init_states.add(HashableNdArray(new_state))

    return [state.array for state in new_init_states]


class ComputeTraceException(Exception):
    pass


def run_model_variable_initial_values(
    *,
    init_state_prototype: Dict[str, str],
    variables: List[str],
    update_fn: Callable,
    equation_system: EquationSystem,
    check_nearby: bool,
) -> Tuple[List, List[np.ndarray], Set[HashableNdArray]]:
    """
    Run the model on possibly *'ed sets of initial conditions

    Parameters
    ----------
    init_state_prototype
        A dictionary of variable names and their initial values {0,1,2,*} where * is a wildcard
    variables
        The variables of the model, in order
    update_fn
    equation_system
    check_nearby
        If true, we include states Hamming distance 1 from the initial state(s).

    Returns
    -------
    List of edges
    """

    # expand any *'ed variables
    def get_states(
        initial_state: Dict[str, str], remaining_variable_states: List[str]
    ) -> List[np.ndarray]:
        if len(remaining_variable_states) == 0:
            return [
                np.array([int(initial_state[var]) for var in variables], dtype=np.int64)
            ]
        else:
            initial_states = []
            for val in range(3):
                resolved_state = initial_state.copy()
                resolved_state[remaining_variable_states[0]] = str(val)
                initial_states += get_states(
                    resolved_state, remaining_variable_states[1:]
                )
            return initial_states

    variable_states = [
        k for k in init_state_prototype if init_state_prototype[k] == "*"
    ]
    try:
        init_states: List[np.ndarray] = get_states(
            init_state_prototype, variable_states
        )
    except ValueError:
        raise ComputeTraceException(
            make_response(error_report("Error constructing initial states"))
        )

    # check to see if we will be overloaded
    if len(variable_states) > MAX_SUPPORTED_VARIABLE_STATES:
        raise ComputeTraceException(
            make_response(
                error_report(
                    f"Web platform is limited to {MAX_SUPPORTED_VARIABLE_STATES} variable states."
                )
            )
        )

    if check_nearby:
        init_states = add_nearby_states(init_states)

    edge_lists: List[List[Edge]] = []
    limit_points: Set[HashableNdArray] = set()
    for state in init_states:
        edge_list, limit_cycle = get_trajectory_edge_list(
            init_state=state,
            update_fn=update_fn,
            target_variables=equation_system.target_variables(),
        )
        edge_lists.append(edge_list)
        limit_points.update(limit_cycle)

    return edge_lists, init_states, limit_points


def compute_trace(
    *,
    model_text: str,
    knockouts: Dict[str, int],
    continuous: Dict[str, bool],
    init_state: Dict[str, str],
    visualize_variables: Dict[str, bool],
    check_nearby: bool,
):
    """Run the cycle finding simulation for an initial state"""

    # create an update function and equation system
    variables, update_fn, equation_system = process_model_text(
        model_text, knockouts, continuous
    )

    # construction of initial values can fail with an exception that contains a response
    try:
        edge_lists, init_states, limit_points = run_model_variable_initial_values(
            init_state_prototype=init_state,
            variables=variables,
            update_fn=update_fn,
            equation_system=equation_system,
            check_nearby=check_nearby,
        )
    except ComputeTraceException as e:
        payload = e.args[0]
        if type(payload) in {str, Response}:
            return payload
        else:
            # should not occur
            return make_response(error_report("Unknown error"))

    ################################
    # give numeric labels to vertices and record their type i.e. initial node, limit cycle, other

    def to_key(edges):
        return frozenset(edges.items())

    class NodeType(IntEnum):
        Other = 0
        InitialNode = 1
        LimitNode = 2

    labels: Dict[frozenset, int] = dict()
    # keys: frozenset of (var, val) pairs defining a point in config space
    # values: integer 'id's for the points of config space

    node_type: Dict[int, NodeType] = dict()
    # keys: integer 'id's for the points of config space
    # values: type of node

    init_states_as_keys = set(
        to_key(dict(zip(variables, state))) for state in init_states
    )
    limit_points_as_keys = set(
        to_key(dict(zip(variables, state.array))) for state in limit_points
    )

    def get_node_type(key) -> NodeType:
        if key in limit_points_as_keys:
            return NodeType.LimitNode
        elif key in init_states_as_keys:
            return NodeType.InitialNode
        else:
            return NodeType.Other

    count = 0
    for edge_list in edge_lists:
        for edge in edge_list:
            source = to_key(edge["source"])
            if source not in labels:
                labels[source] = count
                node_type[count] = get_node_type(source)
                count += 1
            target = to_key(edge["target"])
            if target not in labels:
                labels[target] = count
                node_type[count] = get_node_type(target)
                count += 1

    return_states = []
    for edge_list in edge_lists:
        return_state = to_key(edge_list[-1]["target"])
        return_states.append(labels[return_state])

    source_labels = [
        [labels[to_key(edge["source"])] for edge in edge_list]
        for edge_list in edge_lists
    ]

    variable_level_plots = plot_variable_levels_for_trajectories(
        edge_lists=edge_lists,
        return_states=return_states,
        source_labels=source_labels,
        variables=variables,
        visualize_variables=visualize_variables,
    )

    # create data for the javascript
    edge_lists_by_labels = [
        [
            {
                "source": labels[to_key(edge["source"])],
                "target": labels[to_key(edge["target"])],
            }
            for edge in edge_list
        ]
        for edge_list in edge_lists
    ]

    num_nodes = len(labels)
    height_percent = min(95, math.ceil(50 + 50 / (1 - math.exp(-num_nodes / 100))))
    width_px = 640
    height_px = math.ceil(320 + 320 / (1 - math.exp(-num_nodes / 100)))

    initial_node_positions = graph_layout(edge_lists_by_labels, height_px, width_px)
    nodes_json = json.dumps(
        [
            {
                "id": str(label),
                "group": node_type[label],
                "x": float(initial_node_positions[label][0]),
                "y": float(initial_node_positions[label][1]),
            }
            for label in labels.values()
        ]
    )

    edge_tuples = set()
    for edge_list in edge_lists:
        edge_tuples.update(
            {
                (labels[to_key(edge["source"])], labels[to_key(edge["target"])])
                for edge in edge_list
                if to_key(edge["source"]) != to_key(edge["target"])
            }
        )
        # `if` blocks self loops, which the force-directed graph freaks out about

    edge_json = json.dumps(
        [
            {"source": str(source), "target": str(target)}
            for (source, target) in edge_tuples
        ]
    )

    # respond with the results-of-computation page
    return make_response(
        render_template(
            "compute-trace.html",
            variables=equation_system.target_variables(),
            num_edge_lists=len(edge_lists),
            trajectories=list(
                zip(
                    edge_lists,
                    return_states,
                    source_labels,
                    map(len, edge_lists),
                    variable_level_plots,
                )
            ),
            nodes=nodes_json,
            height_percent=height_percent,
            width_px=width_px,
            height_px=height_px,
            links=edge_json,
        )
    )


def plot_variable_levels_for_trajectories(
    *,
    edge_lists,
    return_states,
    source_labels,
    variables: List[str],
    visualize_variables: Dict[str, bool],
):
    # mask for which variables to viz
    visualize_variable_mask = np.array(
        [visualize_variables[var] for var in variables], dtype=bool
    )

    if np.sum(visualize_variable_mask) <= 0:
        # no viz requested
        return [""] * len(edge_lists)

    # draw a visualization of variable levels
    variable_level_plots = []
    for n, edge_list in enumerate(edge_lists):
        plt.rcParams["svg.fonttype"] = "none"
        plt.figure(figsize=(6, 2))

        # prepare to plot
        data = np.array(
            [
                [edge["source"][variable] for edge in edge_list]
                + [edge_list[-1]["target"][variable]]
                for variable in variables
            ]
        ).T

        # plot
        plt.plot(data[:, visualize_variable_mask])

        # note the repeat region
        if return_states[n] in source_labels[n]:
            plt.axvline(
                source_labels[n].index(return_states[n]), color="k", linestyle="dotted"
            )
        plt.axvline(len(edge_list), color="k", linestyle="dotted")

        # limits and ticks
        plt.ylim([-0.25, 2.25])
        plt.yticks([0, 1, 2])
        plt.xlim([-0.25, len(edge_list) + 0.25])
        plt.xticks(np.arange(len(edge_list) + 1))
        plt.gca().set_xticklabels(source_labels[n] + [return_states[n]])

        # legend and labels
        plt.legend(
            np.array(variables)[visualize_variable_mask],
            bbox_to_anchor=(1.04, 1),
            loc="center left",
        )
        plt.ylabel("Level")
        plt.xlabel("State")
        plt.tight_layout()
        image_filename = f"levels{n}.svg"
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            tmp_file = Path(tmp_dir_name) / image_filename
            plt.savefig(tmp_file, transparent=True, pad_inches=0.0)
            plt.close()
            with open(tmp_file, "r") as image:
                variable_level_plots.append(Markup(image.read()))
    return variable_level_plots



import shutil
import subprocess
import tempfile

from flask import make_response

from steady_cell_phenotype._util import *


def compute_fixed_points(knockout_model, variables, continuous):
    """Run the fixed-point finding computation"""

    # check to make sure that we have macaulay installed
    macaulay_executable = shutil.which("M2")
    if macaulay_executable is None:
        return make_response(
            error_report(
                "Macaulay2 was not found on the server; we cannot do as you ask."
            )
        )

    # we are set to do the computation
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        with open(tmp_dir_name + "/model.txt", "w") as model_file:
            model_file.write(knockout_model)
            model_file.write("\n")

        # convert the model to polynomials
        non_continuous_vars = [
            variable for variable in variables if not continuous[variable]
        ]
        if len(non_continuous_vars) > 0:
            continuity_params = ["-c", "-comit"] + [
                variable for variable in variables if not continuous[variable]
            ]
        else:
            continuity_params = ["-c"]

        convert_to_poly_process = subprocess.run(
            [
                get_resource_path("scp_converter.py"),
                "-n",
                "-i",
                tmp_dir_name + "/model.txt",
                "-o",
                tmp_dir_name + "/model-polys.txt",
            ]
            + continuity_params,
            capture_output=True,
        )

        if convert_to_poly_process.returncode != 0:
            return make_response(
                error_report(
                    "Error running converter!\n{}\n{}".format(
                        html_encode(convert_to_poly_process.stdout),
                        html_encode(convert_to_poly_process.stderr),
                    )
                )
            )

        template_contents = get_text_resource("find_steady_states.m2-template")
        with open(tmp_dir_name + "/find_steady_states.m2", "w") as macaulay_script:
            macaulay_script.write(
                template_contents.format(
                    polynomial_file=tmp_dir_name + "/model-polys.txt",
                    output_file=tmp_dir_name + "/results.txt",
                )
            )

        # TODO: put a limit on the amount of time this can run
        find_steady_states_process = subprocess.run(
            [macaulay_executable, "--script", tmp_dir_name + "/find_steady_states.m2"],
            capture_output=True,
        )
        if find_steady_states_process.returncode != 0:
            # this string appears when there are no solutions. I don't think it appears
            # otherwise.
            if "-infinity (of class InfiniteNumber)" in str(
                find_steady_states_process.stderr
            ):
                return make_response(
                    message(
                        "Macaulay could not find any fixed points, "
                        "this is likely because there aren't any."
                    )
                )
            else:
                with open(tmp_dir_name + "/model-polys.txt", "r") as poly_file:
                    return make_response(
                        error_report(
                            "Error running Macaulay!\n<br>\n{}\n<br>\n{}\n<br>\n{}".format(
                                html_encode(find_steady_states_process.stdout),
                                html_encode(find_steady_states_process.stderr),
                                html_encode(poly_file.read()),
                            )
                        )
                    )

        def process_line(line):
            line = line.strip()
            if len(line) < 2:  # should at least have {}
                return None
            elif line[0] != "{" or line[-1] != "}":
                raise RuntimeError("Malformed response from Macaulay")

            line = line[1:-1]  # strip the {}'s
            try:
                results = map(int, line.split(","))
            except ValueError:
                raise RuntimeError("Malformed response from Macaulay")

            # make sure they are in 0,1,2
            results = map(lambda n: (n + 3) % 3, results)

            return tuple(results)

        try:
            with open(tmp_dir_name + "/results.txt", "r") as file:
                fixed_points = [process_line(line) for line in file]
        except (IOError, RuntimeError):
            return make_response(error_report("Malformed response from Macaulay!"))

    # respond with the results-of-computation page
    return make_response(
        render_template(
            "compute-fixed-points.html", variables=variables, fixed_points=fixed_points
        )
    )



import html
import importlib.resources
import itertools
import pathlib
import re
import string
from typing import Callable, Dict, Iterator, List, Tuple, Union

import numba
import numpy as np
from attr import attrib, attrs
from flask import render_template

from steady_cell_phenotype.equation_system import EquationSystem

MAX_SUPPORTED_VARIABLE_STATES = 6


def get_resource_path(filename: str) -> pathlib.Path:
    """
    Obtain the location of steady_cell_phenotype's data files regardless of package location.

    Parameters
    ----------
    filename: str
        File to obtain
    Returns
    -------
    File's contents as a string
    """
    # See: https://docs.python.org/3.7/library/importlib.html#module-importlib.resources
    with importlib.resources.path("steady_cell_phenotype", filename) as path:
        return path


def get_text_resource(filename: str) -> str:
    """
    Obtain the contents of steady_cell_phenotype's data files regardless of package location.

    Parameters
    ----------
    filename: str
        File to obtain
    Returns
    -------
    File's contents as a string
    """
    # See: https://docs.python.org/3.7/library/importlib.html#module-importlib.resources
    with importlib.resources.open_text(
        "steady_cell_phenotype", filename
    ) as file_handle:
        data = file_handle.read()
    return data


def html_encode(msg):
    if type(msg) is not str:
        msg = msg.decode()
    return html.escape(msg).replace("\n", "<br>").replace(" ", "&nbsp")


def error_report(error_string):
    """display error reports from invalid user input"""
    return render_template("error.html", error_message=error_string)


def message(error_string):
    """display error reports from invalid user input"""
    return render_template("message.html", message=error_string)


def get_model_variables(model) -> Tuple[List[str], List[str]]:
    variables: List[str] = []
    right_sides: List[str] = []
    too_many_eq_msg = (
        "Count of ='s on line {lineno} was {eq_count} but each"
        " line must have a single = sign."
    )
    zero_len_var_msg = "No variable found before = on line {lineno}."
    zero_len_rhs_msg = "No right hand side of equation on line {lineno}."
    invalid_var_name_msg = (
        "One line {lineno}, variable name must be alpha-numeric (plus underscore), "
        "and begin with a letter."
    )
    for lineno, line in enumerate(model.splitlines(), start=1):
        # check for _one_ equals sign
        if line.count("=") != 1:
            raise Exception(
                too_many_eq_msg.format(lineno=lineno, eq_count=line.count("="))
            )
        variable: str
        rhs: str
        variable, rhs = line.split("=")
        variable = variable.strip()
        rhs = rhs.strip()
        # check to see if lhs is a valid symbol.
        # TODO: what variable names does scp_converter.py allow?
        if len(variable) == 0:
            raise Exception(zero_len_var_msg.format(lineno=lineno))
        if not re.match(r"^\w+$", variable) or variable[0] not in string.ascii_letters:
            raise Exception(invalid_var_name_msg.format(lineno=lineno))
        variables.append(variable)
        # do _minimal_ checking on RHS
        if len(rhs) == 0:
            raise Exception(zero_len_rhs_msg.format(lineno=lineno))
        right_sides.append(rhs)
    return variables, right_sides


def decode_int(coded_value, num_variables):
    """Decode long-form int into trinary"""
    if isinstance(coded_value, str):
        if coded_value[:2] == "0x":
            coded_value = int(coded_value, 16)
        else:
            coded_value = int(coded_value)
    exploded_values = []
    for _ in range(num_variables):
        next_value = coded_value % 3
        exploded_values.append(next_value)
        coded_value = (coded_value - next_value) // 3
    exploded_values.reverse()
    return exploded_values


@attrs
class StreamingStats(object):
    """
    Implements Welford's online algorithm for mean and variance calculation. See Knuth Vol 2, pg 232
    """

    mean: float = attrib(init=False, default=float("nan"))
    scaled_var: float = attrib(init=False, default=float("nan"))
    var: float = attrib(init=False, default=float("nan"))
    count: int = attrib(init=False, default=0)

    def add(self, datum: float):
        self.count += 1
        if self.count == 1:
            self.mean = datum
            self.scaled_var = 0
            self.var = 0
        else:
            old_mean = self.mean
            self.mean += (datum - old_mean) / self.count
            self.scaled_var += (datum - old_mean) * (datum - self.mean)

        if self.count > 1:
            self.var = self.scaled_var / (self.count - 1)
        else:
            self.var = 0.0


@attrs
class BinCounter(object):
    """
    Utility class for streaming bincounts.
    """

    bins: np.ndarray = attrib(init=False, default=np.zeros(0, dtype=int))
    max: int = attrib(init=False, default=-1)

    def total(self) -> int:
        return int(np.sum(self.bins))

    def add(self, datum: Union[int, "BinCounter"]):
        if isinstance(datum, int):
            if datum >= len(self.bins):
                old_bins = self.bins
                self.bins = np.zeros(1 + int(1.5 * datum), dtype=int)
                self.bins[: len(old_bins)] = old_bins
            self.bins[datum] += 1
            self.max = max(self.max, datum)
        elif isinstance(datum, BinCounter):
            new_max = max(self.max, datum.max)
            new_bins = np.zeros(shape=(new_max + 1))
            new_bins[: self.max + 1] = self.bins[: self.max + 1]
            new_bins[: datum.max + 1] += datum.bins[: datum.max + 1]
            self.max = new_max
            self.bins = new_bins
        else:
            raise RuntimeError("Cannot handle this.")

    def trimmed_bins(self):
        return np.trim_zeros(filt=self.bins, trim="b")


@numba.njit
def ternary_hash(arr: np.ndarray) -> int:
    accumulator: int = 0
    for idx in range(len(arr)):
        accumulator = 3 * accumulator + arr[idx]
    return int(accumulator)


class HashableNdArray(object):
    """
    A wrapper to make numpy based state arrays hashable
    """

    array: np.ndarray
    hash: int

    def __init__(self, array):
        self.array = array
        # view the array as a ternary expansion; a bijective hash
        self.hash = ternary_hash(array)

    def __repr__(self):
        # return 'HashableNdArray(array=' + repr(self.array) + ')'
        return repr(self.array)

    def __str__(self):
        return str(self.array)

    def __getitem__(self, item):
        return self.array[item]

    def __eq__(self, other: "HashableNdArray"):
        # decided on array_equiv rather than array_equal, because I don't want to worry about
        # the case where the shape is (1,n) or (n,1) and being compared to (n,) or some
        # other such nonsense.
        #
        # note: as this is a performance critical point, I am not going to check the type of 'other'
        # and instead trust that I don't do anything brain-dead.
        return self.hash == other.hash and np.array_equiv(self.array, other.array)

    def __hash__(self):
        return self.hash


def complete_search_generator(
    variables: List[str], constants_vals: Dict[str, int]
) -> Iterator[np.ndarray]:
    """
    Generator which yields all possible states, with constant variables fixed

    Returns
    -------
    Iterator[np.ndarray]
    """
    constants = tuple(constants_vals.keys())
    num_non_constant_vars = len(variables) - len(constants)
    non_constant_vars = list(set(variables) - set(constants))
    for var_dict in map(
        lambda tup: dict(zip(tup[0], tup[1])),
        zip(
            itertools.repeat(non_constant_vars),
            itertools.product(range(3), repeat=num_non_constant_vars),
        ),
    ):
        state = np.fromiter(
            (
                var_dict[var] if var in var_dict else constants_vals[var]
                for var in variables
            ),
            dtype=np.int64,
        )
        yield state


def random_search_generator(
    num_iterations,
    variables: List[str],
    constants_vals: Dict[str, int],
    batch_size=2000,
) -> Iterator[np.ndarray]:
    """
    Generates `num_iterations` random states, with constant variables fixed.

    Parameters
    ----------
    num_iterations : int
        number of iterations to generate
    variables : List[str]
        ordered list of variables
    constants_vals : Dict[str, int]
        dictionary of variables with their constant iteration
    batch_size : int
        size of internal batch, for performance tuning, no other external effect

    Returns
    -------
    Iterator[np.ndarray]
    """
    num_variables = len(variables)
    constant_indices = np.array(
        [idx for idx, var in enumerate(variables) if var in constants_vals],
        dtype=np.int64,
    )
    constant_val_arr = np.array(
        [constants_vals[var] for var in variables if var in constants_vals],
        dtype=np.int64,
    )
    count = 0
    while count < num_iterations:
        init_states = np.random.randint(3, size=(batch_size, num_variables), dtype=int)
        # set constant values
        init_states[:, constant_indices] = constant_val_arr

        for idx in range(batch_size):
            yield init_states[idx, :]
            count += 1
            if count >= num_iterations:
                return


def batcher(
    state_gen: Iterator[np.ndarray], variables, batch_size
) -> Iterator[np.ndarray]:
    num_variables = len(variables)
    batch: np.ndarray
    try:
        while True:
            batch = np.zeros(shape=(batch_size, num_variables), dtype=int)
            for idx in range(batch_size):
                batch[idx] = next(state_gen)
            yield batch
    except StopIteration:
        # noinspection PyUnboundLocalVariable
        yield batch[:idx, :]


def get_phased_trajectory(
    init_state: np.ndarray, update_fn: Callable
) -> Tuple[np.ndarray, HashableNdArray]:
    """
    evolve an initial state until it reaches a limit cycle

    Parameters
    ----------
    init_state
    update_fn

    Returns
    -------
    trajectory, phase-point pair
    """
    state = init_state
    trajectory = list()
    trajectory_set = set()  # set lookup should be faster

    # compute state by state until we have a repeat
    hashable_state = HashableNdArray(state)
    while hashable_state not in trajectory_set:
        trajectory.append(hashable_state)
        trajectory_set.add(hashable_state)
        state = update_fn(state)
        hashable_state = HashableNdArray(state)

    # separate trajectory into in-bound and limit-cycle parts
    repeated_state = HashableNdArray(state)
    repeated_state_index = trajectory.index(repeated_state)
    limit_cycle = trajectory[repeated_state_index:]

    # find state in limit cycle with smallest hash (i.e. smallest lexicographic
    # ordering if there is no integer overflow)
    # this is our phase fixing point
    cycle_min_index: int = 0
    cycle_min: int = hash(limit_cycle[0])
    for idx in range(1, len(limit_cycle)):
        nxt_hash: int = hash(limit_cycle[idx])
        if nxt_hash < cycle_min:
            cycle_min_index = idx
            cycle_min = nxt_hash

    # get trajectory with phase
    phase_idx: int = len(trajectory) - len(limit_cycle) + cycle_min_index
    phased_trajectory = np.array(
        [hashable.array for hashable in trajectory[:phase_idx]], dtype=np.int64
    )

    return phased_trajectory, trajectory[phase_idx]


def get_trajectory(
    init_state: np.ndarray, update_fn: Callable
) -> Tuple[np.ndarray, np.ndarray]:
    """
    evolve an initial state until it reaches a limit cycle

    Parameters
    ----------
    init_state
    update_fn

    Returns
    -------
    trajectory, limit cycle
    """
    state = init_state
    trajectory = list()
    trajectory_set = set()  # set lookup should be faster

    # compute state by state until we have a repeat
    hashable_state = HashableNdArray(state)
    while hashable_state not in trajectory_set:
        trajectory.append(hashable_state)
        trajectory_set.add(hashable_state)
        state = update_fn(state)
        hashable_state = HashableNdArray(state)

    # separate trajectory into in-bound and limit-cycle parts
    repeated_state = HashableNdArray(state)
    repeated_state_index = trajectory.index(repeated_state)

    trimmed_trajectory = np.array(
        [hashable.array for hashable in trajectory[:repeated_state_index]]
    )
    limit_cycle = np.array(
        [hashable.array for hashable in trajectory[repeated_state_index:]]
    )

    return trimmed_trajectory, limit_cycle


def process_model_text(
    model_text: str, knockouts: Dict[str, int], continuous: Dict[str, bool]
) -> Tuple[List[str], Callable, EquationSystem]:
    equation_system = EquationSystem.from_text(model_text)
    equation_system = equation_system.continuous_functional_system(
        continuous_vars=tuple([var for var in continuous if continuous[var]])
    )
    equation_system = equation_system.knockout_system(knockouts)

    # create an update function
    variables: List[str]
    update_fn: Callable
    variables, update_fn = equation_system.as_numpy()

    return variables, update_fn, equation_system



from __future__ import annotations

import operator
from copy import deepcopy
from functools import partial, reduce
from html import escape
# noinspection PyUnresolvedReferences
from math import floor
from typing import Callable, Optional, Sequence

# noinspection PyProtectedMember
from bs4 import ResultSet

from steady_cell_phenotype.poly import *

UNIVARIATE_FUNCTIONS = ["NOT"]
BIVARIATE_FUNCTIONS = ["MAX", "MIN", "CONT"]

DEFAULT_COMPARTMENT_NAME = "compartment1"

try:
    # noinspection PyUnresolvedReferences
    import multiprocessing

    PARALLEL = True
except ImportError:
    PARALLEL = False


####################################################################################################


class ParseError(Exception):
    """Exception raised when we go down a wrong path in parsing

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


####################################################################################################


def tokenize(input_string: str) -> Sequence[Tuple[str, Union[str, int]]]:
    """This tokenizer takes a string and tokenizes it into a list of 2-tuples. Each
    tuple is of the form (TOKEN_TYPE, VALUE) where TOKEN_TYPE can be one of OPEN_PAREN,
    CLOSE_PAREN, COMMA, EQUALS, PLUS, MINUS, TIMES, EXP, FUNCTION, or SYMBOL. VALUE only contains
    non-redundant information in the case of SYMBOL, in which is contains the string
    representation of the symbol.

    This method can fail for invalid equations, throwing an exception.

    """
    tokenized_list = []
    whitespace = [" ", "\t"]
    punctuation = ["(", ")", ",", "=", "+", "-", "*", "^"]
    function_names = ["MAX", "MIN", "NOT"]
    while len(input_string) > 0:
        if input_string[0] in whitespace:
            # remove whitespace: spaces and tabs
            input_string = input_string[1:]
        elif input_string[0] == "(":
            tokenized_list.append(("OPEN_PAREN", "("))
            input_string = input_string[1:]
        elif input_string[0] == ")":
            tokenized_list.append(("CLOSE_PAREN", ")"))
            input_string = input_string[1:]
        elif input_string[0] == ",":
            tokenized_list.append(("COMMA", ","))
            input_string = input_string[1:]
        elif input_string[0] == "=":
            tokenized_list.append(("EQUALS", "="))
            input_string = input_string[1:]
        elif input_string[0] == "+":
            tokenized_list.append(("PLUS", "+"))
            input_string = input_string[1:]
        elif input_string[0] == "-":
            tokenized_list.append(("MINUS", "-"))
            input_string = input_string[1:]
        elif input_string[0] == "*":
            tokenized_list.append(("TIMES", "*"))
            input_string = input_string[1:]
        elif input_string[0] == "^":
            tokenized_list.append(("EXP", "^"))
            input_string = input_string[1:]
        elif input_string[0:3].upper() in function_names:
            tokenized_list.append(("FUNCTION", input_string[0:3].upper()))
            input_string = input_string[3:]
        else:
            # must be a name or constant. can be of variable length, terminated by punctuation or
            # whitespace
            index = 0
            while (
                index < len(input_string)
                and not input_string[index] in punctuation
                and not input_string[index] in whitespace
            ):
                index += 1
            if index > 0:
                try:
                    # check to see if this is a constant.
                    const = int(input_string[0:index])
                    tokenized_list.append(("CONSTANT", const))
                except ValueError:
                    # if it isn't parsable as an int, it is a symbol
                    tokenized_list.append(("SYMBOL", input_string[0:index]))
                input_string = input_string[index:]
            else:
                raise Exception("Error in tokenization, cannot understand what this is")
    return tokenized_list


####################################################################################################


def find_all(find_token_string, formula):
    """generator which yields all indices (in order) of occurrences of a particular type
    (find_token_string) of token in formula, which is an enumerable of (token,value) pairs

    will also work if find_token_string is a list of strings
    """
    if type(find_token_string) == str:
        for n, (token_string, _) in enumerate(formula):
            if token_string == find_token_string:
                yield n
    else:
        # will work with lists, but we don't check it
        for n, (token_string, _) in enumerate(formula):
            if token_string in find_token_string:
                yield n


####################################################################################################
# recursive descent parser
####################################################################################################


def translate_to_expression(
    formula: Sequence[Tuple[str, Union[str, int]]]
) -> ExpressionOrInt:
    """
    Parse formula using recursive descent.

    Recurses left-to-right, building the formula on the way back up.

    Parameters
    ----------
    formula
        tokenized list

    Returns
    -------
    ExpressionOrInt
    """
    if len(formula) <= 0:
        raise ParseError("can't parse an empty formula")

    next_token_type, next_token_content = formula[0]

    # try to parse addition and subtraction
    for additive_index in find_all(["PLUS", "MINUS"], formula):
        try:
            argument_one = translate_to_expression(formula[0:additive_index])
            argument_two = translate_to_expression(formula[additive_index + 1 :])

            # we succeed!
            if formula[additive_index][0] == "PLUS":
                return argument_one + argument_two
            else:
                assert formula[additive_index][0] == "MINUS"
                return (
                    argument_one - argument_two
                )  # Note: This is, in some sense, wrong, but fixed elsewhere. TODO: Remove this
        except ParseError:
            pass

    # try to parse as a unary minus: -ARG
    if next_token_type == "MINUS":
        try:
            argument = translate_to_expression(formula[1:])
            # we succeed!
            return -argument
        except ParseError:
            pass

    # try to parse as a unary plus: +ARG
    if next_token_type == "PLUS":
        try:
            argument = translate_to_expression(formula[1:])
            # we succeed!
            return argument
        except ParseError:
            pass

    # try to parse multiplication
    for mult_index in find_all("TIMES", formula):
        try:
            argument_one = translate_to_expression(formula[0:mult_index])
            argument_two = translate_to_expression(formula[mult_index + 1 :])
            # we succeed!
            return argument_one * argument_two
        except ParseError:
            pass

    # try to parse exponentiation --- we only allow atomic integer constants in the exponent because
    # these are not to be taken mod 3
    if len(formula) >= 3 and formula[-2][0] == "EXP" and formula[-1][0] == "CONSTANT":
        exp_index = len(formula) - 2
        try:
            base = translate_to_expression(formula[0:exp_index])
            # again, we only accept constants in the exponent!
            if formula[-1][0] != "CONSTANT":
                raise ParseError("must have constant in the exponent")
            exponent = formula[-1][1]
            # we succeed!
            return base ** exponent
        except ParseError:
            pass

    # try to parse parentheses: (ARG)
    if next_token_type == "OPEN_PAREN" and formula[-1][0] == "CLOSE_PAREN":
        try:
            argument = translate_to_expression(formula[1:-1])
            # we succeed!
            return argument
        except ParseError:
            pass

    # try to parse CONST/SYMB
    if len(formula) == 1:
        if next_token_type == "CONSTANT":
            constant: int = next_token_content
            return constant
        elif next_token_type == "SYMBOL":
            symbol_text = next_token_content
            return Monomial({symbol_text: 1})
        else:
            raise ParseError("Invalid singleton")

    # try to parse functions
    if next_token_type == "FUNCTION":
        function_name = next_token_content
        # formatting sanity check; must have FUNCTION OPEN_PAREN ... CLOSE_PAREN
        if (
            len(formula) < 2
            or formula[1][0] != "OPEN_PAREN"
            or formula[-1][0] != "CLOSE_PAREN"
        ):
            raise ParseError(
                "Parse error, must have form FUNCTION OPEN_PAREN ... CLOSE_PAREN"
            )

        # lop off the function name and parentheses
        inner_formula = formula[2:-1]

        if function_name in UNIVARIATE_FUNCTIONS:
            argument = translate_to_expression(
                inner_formula
            )  # may throw ParseError, which we do not catch here
            # we succeed!
            return Function(function_name, [argument])
        elif function_name in BIVARIATE_FUNCTIONS:
            # try to find the comma
            for comma_index in find_all("COMMA", inner_formula):
                try:
                    argument_one = translate_to_expression(inner_formula[0:comma_index])
                    argument_two = translate_to_expression(
                        inner_formula[comma_index + 1 :]
                    )
                    # if we succeed, return the function
                    return Function(function_name, [argument_one, argument_two])
                except ParseError:
                    # try another comma
                    pass
            # no comma gave a valid parse
            raise ParseError("could not parse function call")
        else:
            # function name was unrecognized
            raise ParseError("This is not a function I recognize!")

    # in the case that nothing worked
    raise ParseError("Parse error!")


################################################################################################
# helper functions for parallelization, b/c Python's pickling is weirdly inflexible


def evaluate_parallel_helper(pair, mapping_dict):
    """
    Because python won't let you use lambdas in multiprocessing
    :param pair:
    :param mapping_dict:
    :return:
    """
    variable_name, expression = pair
    if isinstance(expression, int) or expression.is_constant():
        return variable_name, expression
    else:
        return variable_name, expression.eval(mapping_dict)


def continuity_helper(
    control_variable: str, equation: ExpressionOrInt, continuous_vars: Sequence[str]
) -> Tuple[str, ExpressionOrInt]:
    """
    Because python won't let you use lambdas in multiprocessing
    :param control_variable:
    :param equation:
    :param continuous_vars:
    :return:
    """
    if control_variable not in continuous_vars:
        return control_variable, equation
    elif isinstance(equation, int) or equation.is_constant():
        return control_variable, int(equation)
    else:
        return control_variable, equation.continuous_polynomial_version(
            control_variable
        )


def polynomial_output_parallel_helper(equation: ExpressionOrInt) -> ExpressionOrInt:
    """
    Converter from an ExpressionOrInt, to one which is a polynomial.

    Required because python won't let you use lambdas in multiprocessing

    Parameters
    ----------
    equation
        The equation to convert

    Returns
    -------
    ExpressionOrInt
    """
    if isinstance(equation, int) or equation.is_constant():
        return int(equation)
    else:
        return equation.as_polynomial()


####################################################################################################


def parse_mathml_to_function(
    input_variables: List[str], mathml_tag: Tag
) -> Callable[..., bool]:
    def parse_mathml_helper(input_variable_list: List[str], mathml_subtag) -> Callable:
        """
        Recursive building of the function from the inner mathml.

        Parameters
        ----------
        input_variable_list
            ordered list of variable names
        mathml_subtag
            inner mathml to parse to a function

        Returns
        -------
        Callable
        """
        tag_type = mathml_subtag.name
        if tag_type == "cn":
            # constant
            if mathml_subtag.attrs["type"] != "integer":
                raise ParseError("Non-integer data type found")
            try:
                value = int(mathml_subtag.text)

                # noinspection PyUnusedLocal
                def f(input_values):
                    return value

                return f
            except ValueError:
                raise ParseError(f"Could not parse {mathml_subtag.text} as an integer")
        if tag_type == "ci":
            # variable
            variable_name = mathml_subtag.text.strip()
            try:
                variable_index = input_variable_list.index(variable_name)
            except ValueError:
                raise ParseError(
                    f"'{variable_name}' not found in list"
                    f" of variables: {input_variable_list}"
                )

            # noinspection PyUnusedLocal
            def f(input_values):
                return input_values[variable_index]

            return f
        if tag_type == "apply":
            tag_contents: List[Tag] = list(
                filter(lambda x: type(x) == Tag, mathml_subtag.contents)
            )
            if len(tag_contents) <= 0:
                raise ParseError("Empty apply")
            operation = tag_contents[0].name
            operand_functions: List[Callable] = [
                parse_mathml_helper(input_variables, operand_tag)
                for operand_tag in tag_contents[1:]
            ]

            # handle 'not', 'unary minus', and the modulus
            unary_operations = {
                "not": operator.not_,
                "minus": operator.neg,
                "floor": lambda x: floor(x),
            }
            if operation in unary_operations and len(operand_functions) == 1:
                operator_function = unary_operations[operation]
                operand = operand_functions[0]

                def f(input_values):
                    return operator_function(operand(input_values))

                return f

            # the commutative operators
            commutative_operations = {
                "max": lambda a, b: max(a, b),
                "min": lambda a, b: min(a, b),
                "plus": operator.add,
                "times": operator.mul,
                "and": operator.and_,
                "or": operator.or_,
                "xor": operator.xor,
                "eq": operator.eq,
                "neq": operator.ne,
            }
            if operation in commutative_operations and len(operand_functions) >= 1:
                operator_function = commutative_operations[operation]

                def f(input_values):
                    return reduce(
                        operator_function,
                        [function(input_values) for function in operand_functions],
                    )

                return f

            # remaining binary operators: comparisons and other non-commutative operations
            binary_operations = {
                "eq": operator.eq,
                "neq": operator.ne,
                "gt": operator.gt,
                "lt": operator.lt,
                "geq": operator.ge,
                "leq": operator.le,
                "minus": operator.sub,
                "divide": operator.truediv,
                "power": operator.pow,
                "rem": operator.mod,
            }
            if operation in binary_operations and len(operand_functions) == 2:
                operator_function = binary_operations[operation]
                operand_a, operand_b = operand_functions

                def f(input_values):
                    return operator_function(
                        operand_a(input_values), operand_b(input_values)
                    )

                return f

            # we didn't handle the parse
            raise ParseError(
                f"Unsupported function: {operation} on {len(operand_functions)} operands"
            )

    # examine the contents of the mathml, there may be strings (probably \n) which we should ignore
    contents: List[Tag] = list(filter(lambda x: type(x) == Tag, mathml_tag.contents))
    if len(contents) != 1:
        raise ParseError

    return parse_mathml_helper(input_variables, contents[0])


def parse_sbml_qual_function(
    input_variables: List[str], function_terms: Tag
) -> Callable[..., int]:
    default_levels: ResultSet = function_terms.findChildren("qual:defaultterm")
    if len(default_levels) != 1:
        raise ParseError("There should be exactly one default level!")
    default_level_tag: Tag = default_levels[0]
    if "qual:resultlevel" not in default_level_tag.attrs:
        raise ParseError("Malformed tags")
    try:
        default_level: int = int(default_level_tag.attrs["qual:resultlevel"])
    except ValueError:
        raise ParseError(
            f"Non-integer default level {default_level_tag.attrs['qual:resultlevel']}"
        )

    conditional_levels: List[Tuple[int, Callable[..., bool]]] = []
    for conditional_level_tag in function_terms.findChildren("qual:functionterm"):
        if "qual:resultlevel" not in conditional_level_tag.attrs:
            raise ParseError("Missing qual:resultlevel")
        try:
            level: int = int(conditional_level_tag.attrs["qual:resultlevel"])
        except ValueError:
            raise ParseError("Non-int qual:resultlevel")

        level_function: Callable[..., bool] = parse_mathml_to_function(
            input_variables, conditional_level_tag.findChild("math")
        )

        conditional_levels.append((level, level_function))

    def function_realization(input_values) -> int:
        for value, function in conditional_levels:
            if function(input_values):
                return value
        return default_level

    return function_realization


####################################################################################################


@attrs(init=False, slots=True, repr=False, str=False)
class EquationSystem(object):
    _formula_symbol_table: List[str] = attrib()
    _equation_dict: Dict[str, ExpressionOrInt] = attrib()
    _lines: List[Tuple[Optional[str], Optional[str]]] = attrib()

    # _lines: Tuples of variable name and comment, both optional, for printing

    def __init__(
        self,
        *,
        formula_symbol_table: List[str] = None,
        equation_dict: Dict[str, ExpressionOrInt] = None,
        lines: List[Tuple[Optional[str], Optional[str]]] = None,
    ):
        if formula_symbol_table is not None and equation_dict is not None:
            self._formula_symbol_table = deepcopy(formula_symbol_table)
            self._equation_dict = deepcopy(equation_dict)
            if lines is None:
                self._lines = [(symbol, None) for symbol in formula_symbol_table]
            else:
                self._lines = deepcopy(lines)
        elif formula_symbol_table is None and equation_dict is None:
            self._formula_symbol_table = []
            self._equation_dict = dict()
            self._lines = list()
        else:
            raise RuntimeError(
                "Must specify either an empty system,"
                " or both formula_symbol_table and equation_dict"
            )

    @staticmethod
    def from_text(lines: str) -> EquationSystem:
        """
        Create a system of equations from text

        Returns
        -------
        EquationSystem
        """
        equation_system: EquationSystem = EquationSystem()
        for line_number, line in enumerate(lines.strip().splitlines()):
            equation_system.parse_and_add_equation(line)
        return equation_system

    @staticmethod
    def from_sbml_qual(xml_string: str) -> Union[EquationSystem, str]:
        """
        Create a system of equations from an SBML-qual string.

        Parameters
        ----------
        xml_string: str
            SBML-qual string representation of equation system

        Returns
        -------
        EquationSystem
            if we can, otherwise an error string
        """
        soup = BeautifulSoup(xml_string, "lxml")

        species_list: List[Tag] = soup.findChildren("qual:qualitativespecies")
        # do some basic checks
        if len(species_list) == 0:
            return "No species defined in the model"
        for species in species_list:
            if "qual:id" not in species.attrs:
                return "There was a species without a name!"
            species_name = species.attrs["qual:id"]
            if len(species_name) == 0 or not species_name[0].isalpha():
                return f"Species name {species_name} is invalid!"

        symbols: List[str] = [species.attrs["qual:id"] for species in species_list]

        # record their 'maxlevel', i.e. if they are boolean or ternary, or whatever.
        # We are only going to support ternary, so error out if something else is encountered
        for species in species_list:
            if "qual:maxlevel" not in species.attrs:
                return (
                    f"No max level provided for {species.attrs['qual:id']}, expecting 2"
                )

            try:
                if int(species.attrs["qual:maxlevel"]) != 2:
                    return (
                        f"Max level provided for {species.attrs['qual:id']}"
                        f" is {species.attrs['qual:maxlevel']}, expecting 2"
                    )
            except ValueError:
                return (
                    f"Max level provided for {species.attrs['qual:id']} is"
                    f" {species.attrs['qual:maxlevel']}, cannot parse as an integer"
                )

        # some of these might be constants
        equations: Dict[str, ExpressionOrInt] = {}
        for species in species_list:
            if (
                "qual:constant" in species.attrs
                and species.attrs["qual:constant"].lower() != "false"
            ):
                if "qual:initiallevel" in species.attrs:
                    # if the initial level is set then we make the variable constant in the
                    # sense that the update function sets that particular value.
                    try:
                        constant_level = int(species.attrs["qual:initiallevel"])
                        equations[species.attrs["qual:id"]] = constant_level
                    except ValueError:
                        return (
                            f"Level provided for {species.attrs['qual:id']} is"
                            f" {species.attrs['qual:initiallevel']}, cannot parse as an integer"
                        )

                else:
                    # if the initial level isn't set then we make the variable constant in the
                    # sense that the update function is the identity.
                    equations[species.attrs["qual:id"]] = Monomial.as_var(
                        species.attrs["qual:id"]
                    )

        # now we parse the transition functions into Expressions
        for transition in soup.findChildren("qual:transition"):
            # get the target
            target_variables = [
                output["qual:qualitativespecies"]
                for output in transition.findChildren("qual:output")
            ]
            if len(target_variables) != 1:
                return "Transition functions must have exactly one output."
            target_variable = target_variables[0]

            if target_variable in equations:
                return (
                    f"Variable {target_variable} appears twice as"
                    f" the target of an update function!"
                )

            input_variables = [
                input_variable["qual:qualitativespecies"]
                for input_variable in transition.findChildren("qual:input")
            ]

            function_terms: ResultSet = transition.findChildren(
                "qual:listoffunctionterms"
            )
            if len(function_terms) != 1:
                return "Should have exactly one list of function terms"
            try:
                transition_function: Callable = parse_sbml_qual_function(
                    input_variables, function_terms[0]
                )
            except ParseError as e:
                return "Could not parse functions " + str(e)

            expression: ExpressionOrInt = 0

            def interpolation_monomial(params):
                input_value, input_variable = params
                return 1 - (Monomial.as_var(input_variable) - input_value) ** 2

            for input_values in product([0, 1, 2], repeat=len(input_variables)):
                output_value: int = transition_function(input_values)
                if output_value != 0:
                    # g(x) = sum_{c\in \F_3^n} h(c) prod_{j=0}^n (1-(x_j-c_j)**2)
                    expression += output_value * reduce(
                        operator.mul,
                        map(interpolation_monomial, zip(input_values, input_variables)),
                    )
            equations[target_variable] = expression

        for variable in symbols:
            if variable not in equations:
                return f"No update function for {variable} specified!"

        lines = [(symbol, None) for symbol in symbols]

        return EquationSystem(
            formula_symbol_table=symbols, equation_dict=equations, lines=lines
        )

    def as_poly_system(self) -> EquationSystem:
        formula_symbol_table = deepcopy(self._formula_symbol_table)
        equation_dict = {
            target_var: eqn.as_polynomial() if isinstance(eqn, Expression) else eqn
            for target_var, eqn in self._equation_dict.items()
        }

        return EquationSystem(
            formula_symbol_table=formula_symbol_table,
            equation_dict=equation_dict,
            lines=deepcopy(self._lines),
        )

    def symbol_table(self):
        return set(self._formula_symbol_table).union(self._equation_dict.keys())

    def formula_symbol_table(self):
        return deepcopy(self._formula_symbol_table)

    def target_variables(self) -> Tuple[str]:
        return tuple(self._equation_dict.keys())

    def variables_that_vary(self) -> List[str]:
        return [
            var
            for var, eqn in self._equation_dict.items()
            if type(eqn) != int and not eqn.is_constant()
        ]

    def constant_variables(self) -> List[str]:
        return [
            var
            for var, eqn in self._equation_dict.items()
            if type(eqn) == int or eqn.is_constant()
        ]

    def consistent(self) -> bool:
        """
        Check if this makes sense as an update function.

        Returns
        -------
        True if the set of formula symbols is contained in the set of target variables
        """
        return set(self._formula_symbol_table) <= set(self._equation_dict.keys())

    ################################################################################################

    def eval(self, state: Dict[str, int]) -> Dict[str, int]:
        if state.keys() != set(self._equation_dict.keys()):
            raise RuntimeError("Evaluating state on incorrect set of variables")
        return {
            variable: formula.eval(state)
            if isinstance(formula, Expression)
            else formula
            for variable, formula in self._equation_dict.items()
        }

    ################################################################################################

    # def as_sympy(self):
    #     variables = tuple(self._equation_dict.keys())
    #
    #     return variables, \
    #            {var: self._equation_dict[var] if is_integer(self._equation_dict[var])
    #            else self._equation_dict[var].as_sympy()
    #            if isinstance(self._equation_dict[var], Expression)
    #            else sympy.Integer(self._equation_dict[var])
    #             for var in self._equation_dict}

    def as_numpy(self) -> Tuple[Tuple[str], Callable]:
        variables: Tuple[str] = tuple(self._equation_dict.keys())

        functions = [
            self._equation_dict[var].as_numpy_str(variables)
            if isinstance(self._equation_dict[var], Expression)
            else str(np.mod(int(self._equation_dict[var]), 3))
            for var in variables
        ]

        function_str = "".join(
            [
                "update_function = lambda state: np.array([",
                ",".join(["np.mod(" + function + ", 3)" for function in functions]),
                "], dtype=np.int64)",
            ]
        )

        # See https://docs.python.org/3/library/functions.html#exec
        locals_dict = dict()
        exec(
            function_str, globals(), locals_dict
        )  # now there is a 'update_function' defined
        update_function = locals_dict["update_function"]
        return variables, update_function

    ################################################################################################

    def as_sbml_qual(self) -> BeautifulSoup:
        soup = BeautifulSoup(features="xml")

        sbml_top_tag: Tag = soup.new_tag(
            "sbml",
            attrs={
                "xmlns": "http://www.sbml.org/sbml/level3/version1/core",
                "level": "3",
                "version": "1",
                "xmlns:qual": "http://www.sbml.org/sbml/level3/version1/qual/version1",
                "qual:required": "true",
            },
        )
        soup.append(sbml_top_tag)

        model = soup.new_tag("model", attrs={"id": "SteadyCellPhenotype model"})
        sbml_top_tag.append(model)

        species_list = soup.new_tag(
            "qual:listOfQualitativeSpecies",
            attrs={
                "xmlns:qual": "http://www.sbml.org/sbml/level3/version1/qual/version1"
            },
        )
        model.append(species_list)

        transitions_list = soup.new_tag(
            "qual:listOfTransitions",
            attrs={
                "xmlns:qual": "http://www.sbml.org/sbml/level3/version1/qual/version1"
            },
        )
        model.append(transitions_list)

        # add species to species list, and transition functions to their own list
        for symbol in self._formula_symbol_table:
            update_formula: ExpressionOrInt = self._equation_dict[symbol]

            if (
                isinstance(update_formula, Expression)
                and not update_formula.is_constant()
            ):
                # add to species list
                symbol_tag = soup.new_tag(
                    "qual:qualitativeSpecies",
                    attrs={
                        "qual:maxLevel": 2,
                        "qual:compartment": DEFAULT_COMPARTMENT_NAME,
                        "qual:id": symbol,
                    },
                )
                species_list.append(symbol_tag)

                transition_tag = soup.new_tag(
                    "qual:transition", attrs={"qual:id": "transition_" + symbol}
                )
                transitions_list.append(transition_tag)

                input_list = soup.new_tag("qual:listOfInputs")
                transition_tag.append(input_list)
                for idx, variable in enumerate(update_formula.get_variable_set()):
                    input_list.append(
                        soup.new_tag(
                            "qual:input",
                            attrs={
                                "qual:qualitativeSpecies": variable,
                                "qual:transitionEffect": "none",
                                "qual:id": f"transition_{symbol}_input_{idx}",
                            },
                        )
                    )

                output_list = soup.new_tag("qual:listOfOutputs")
                transition_tag.append(output_list)
                output_list.append(
                    soup.new_tag(
                        "qual:output",
                        attrs={
                            "qual:qualitativeSpecies": symbol,
                            "qual:transitionEffect": "assignmentLevel",
                            "qual:id": f"transition_{symbol}_output",
                        },
                    )
                )

                function_terms = soup.new_tag("qual:listOfFunctionTerms")
                transition_tag.append(function_terms)

                function_terms.append(
                    soup.new_tag("qual:defaultTerm", attrs={"qual:resultLevel": 0})
                )

                level1_function_term = soup.new_tag(
                    "qual:functionTerm", attrs={"qual:resultLevel": 1}
                )
                level1_function_term.append(update_formula.as_sbml_qual_relation(1))
                function_terms.append(level1_function_term)

                level2_function_term = soup.new_tag(
                    "qual:functionTerm", attrs={"qual:resultLevel": 2}
                )
                level2_function_term.append(update_formula.as_sbml_qual_relation(2))
                function_terms.append(level2_function_term)

            else:
                # set constant and initial level flags
                update_formula = int(update_formula)
                symbol_tag = soup.new_tag(
                    "qual:qualitativeSpecies",
                    attrs={
                        "qual:maxLevel": 2,
                        "qual:compartment": DEFAULT_COMPARTMENT_NAME,
                        "qual:id": symbol,
                        "qual:constant": "true",
                        "qual:initialLevel": str(update_formula),
                    },
                )
                species_list.append(symbol_tag)

        compartments_list = soup.new_tag("listOfCompartments")
        model.append(compartments_list)
        compartments_list.append(
            soup.new_tag(
                "compartment",
                attrs={"constant": "true", "id": DEFAULT_COMPARTMENT_NAME},
            )
        )

        return soup

    ################################################################################################

    def __getitem__(self, item):
        return self._equation_dict[item]

    ################################################################################################

    def continuous_polynomial_system(
        self, continuous_vars: Sequence[str] = None
    ) -> EquationSystem:
        """
        Get continuous version of system

        Parameters
        ----------
        continuous_vars sequence of variable names. if not specified, all are made continuous

        Returns
        -------
        EquationSystem
        """
        if continuous_vars is None:
            continuous_vars = list(self._equation_dict.keys())

        if PARALLEL:
            with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
                continuous_equations = dict(
                    pool.starmap(
                        continuity_helper,
                        [
                            (var, eqn, continuous_vars)
                            for var, eqn in self._equation_dict.items()
                        ],
                    )
                )
        else:
            continuous_equations = {
                control_variable: continuity_helper(
                    control_variable, equation, continuous_vars
                )
                for control_variable, equation in self._equation_dict.items()
            }
        return EquationSystem(
            formula_symbol_table=self._formula_symbol_table,
            equation_dict=continuous_equations,
            lines=self._lines,
        )

    ################################################################################################

    def continuous_functional_system(
        self, continuous_vars: Sequence[str] = None
    ) -> EquationSystem:
        """
        Get continuous version of system

        Parameters
        ----------
        continuous_vars sequence of variable names. if not specified, all are made continuous

        Returns
        -------
        EquationSystem
        """
        if continuous_vars is None:
            continuous_vars = set(self._equation_dict.keys())

        continuous_vars = set(continuous_vars)

        continuous_equations = {
            var: Function("CONT", [Monomial.as_var(var), expr])
            if var in continuous_vars
            else expr
            for var, expr in self._equation_dict.items()
        }

        return EquationSystem(
            formula_symbol_table=self._formula_symbol_table,
            equation_dict=continuous_equations,
            lines=self._lines,
        )

    ################################################################################################

    def simplify(self) -> EquationSystem:
        """
        Create simplified system which propagates constant variables to formulas where used

        Returns
        -------
        EquationSystem
        """

        constant_variables = self.constant_variables()
        constant_dict: Dict[str, int] = dict()
        for constant_variable in constant_variables:
            constant = int(self._equation_dict[constant_variable])
            constant_dict[constant_variable] = constant

        equations = {
            var: eqn.eval(constant_dict) if isinstance(eqn, Expression) else eqn
            for var, eqn in self._equation_dict.items()
        }

        return EquationSystem(
            formula_symbol_table=self._formula_symbol_table,
            equation_dict=equations,
            lines=self._lines,
        )

    ################################################################################################

    def knockout_system(self, knockouts: Dict[str, int]) -> EquationSystem:
        """
        Create 'knockout' system, where equations for certain variables are set to constants

        Parameters
        ----------
        knockouts: Dict[str, int] format is variable-name: value

        Returns
        -------
        EquationSystem
        """
        equation_dict = deepcopy(self._equation_dict)
        for var, val in knockouts.items():
            assert var in equation_dict, "Can't knockout a variable which isn't present"
            equation_dict[var] = int(val) % 3

        return EquationSystem(
            formula_symbol_table=self._formula_symbol_table,
            equation_dict=equation_dict,
            lines=self._lines,
        )

    ################################################################################################

    def compose(self, other: EquationSystem) -> EquationSystem:
        assert set(self._equation_dict.keys()) == set(
            other._equation_dict.keys()
        ), "incompatible systems!"

        if PARALLEL:
            with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
                # start mapping via eval
                evaluator = partial(
                    evaluate_parallel_helper, mapping_dict=other._equation_dict
                )
                composed_dict = dict(
                    pool.map(evaluator, tuple(self._equation_dict.items()))
                )
        else:
            composed_dict = {
                target_var: eqn.eval(other._equation_dict)
                for target_var, eqn in self._equation_dict
            }

        return EquationSystem(
            formula_symbol_table=deepcopy(self._formula_symbol_table),
            equation_dict=composed_dict,
        )

    def self_compose(self, count: int) -> EquationSystem:
        assert count >= 0, "negative powers unsupported!"

        if count == 0:
            # every variable maps to itself in the identity map.
            return EquationSystem(
                formula_symbol_table=self._formula_symbol_table,
                equation_dict={
                    var: Monomial.as_var(var) for var in self._equation_dict.keys()
                },
            )
        elif count == 1:
            return self
        elif count % 2 == 0:
            square_root_system = self.self_compose(count // 2)
            return square_root_system.compose(square_root_system)
        else:  # count % 2 == 1
            pseudo_square_root_system = self.self_compose(count // 2)
            return pseudo_square_root_system.compose(pseudo_square_root_system).compose(
                self
            )

    ################################################################################################

    def __str__(self):
        if len(self._equation_dict) == 0:
            return "Empty System"
        else:
            return "\n".join(
                [
                    (
                        str(var)
                        + "="
                        + str(self._equation_dict[var])
                        + (" # " + escape(comment) if comment is not None else "")
                    )
                    if var is not None
                    else ("# " + escape(comment) if comment is not None else "")
                    for var, comment in self._lines
                ]
            )

    __repr__ = __str__

    ################################################################################################

    def __iter__(self):
        return self._equation_dict.items().__iter__()

    ################################################################################################

    def polynomial_output(self, translate_symbol_names=True) -> str:
        if PARALLEL:
            with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
                equations_as_polys = list(
                    pool.map(
                        polynomial_output_parallel_helper, self._equation_dict.values()
                    )
                )
        else:
            equations_as_polys = [
                equation.as_polynomial()
                if isinstance(equation, Expression)
                else equation
                for equation in self._equation_dict.values()
            ]

        if translate_symbol_names:
            translation_names = {
                var: "x" + str(idx) for idx, var in enumerate(self.target_variables())
            }
            return "\n".join(
                [
                    "f"
                    + str(idx)
                    + "="
                    + (
                        str(eqn.rename_variables(translation_names))
                        if isinstance(eqn, Expression)
                        else str(eqn)
                    )
                    for idx, eqn in enumerate(equations_as_polys)
                ]
            )
        else:
            return "\n".join(
                [
                    str(target) + "=" + str(eqn)
                    for target, eqn in zip(
                        self._equation_dict.keys(), equations_as_polys
                    )
                ]
            )

    ################################################################################################

    def parse_and_add_equation(self, line: str):
        """
        Parse a line of the form 'symbol=formula' and add it to the system.

        Parameters
        ----------
        line: str
            A line of the form 'symbol=formula # comment' or pure comment starting with '#'

        Returns
        -------
        None
        """
        line = line.strip()

        comment_idx = line.find("#")
        comment: Optional[str] = None
        if comment_idx != -1:
            comment = line[comment_idx + 1 :].strip()
            line = line[:comment_idx].strip()

        if len(line) == 0:
            self._lines.append((None, comment))
            return  # empty or comment

        if len(line.splitlines()) != 1:
            raise ParseError("This function does not handle multi-line equations.")

        line = line.replace("-", "+(-1)*")  # parsing aide

        tokenized_list = tokenize(line)

        # Formulas should be of the form 'symbol=function'
        if (
            len(tokenized_list) < 3
            or tokenized_list[0][0] != "SYMBOL"
            or tokenized_list[1][0] != "EQUALS"
        ):
            raise ParseError("Formula did not begin with a symbol then equals sign!")
        target_variable = tokenized_list[0][1]

        # the rest should correspond to the formula
        tokenized_formula = tokenized_list[2:]

        # gather the symbols from the function, which may not all be new
        symbols = [
            symbol for (type_str, symbol) in tokenized_formula if type_str == "SYMBOL"
        ]

        equation = translate_to_expression(tokenized_formula)

        # if no errors, make updates
        for symb in symbols:
            if symb not in self._formula_symbol_table:
                self._formula_symbol_table.append(symb)
        self._equation_dict[target_variable] = equation

        self._lines.append((target_variable, comment))

    ################################################################################################



#!/usr/bin/env python3
# Copyright Adam C. Knapp 2018-2021
# License: CC-BY 4.0 https://creativecommons.org/licenses/by/4.0/
#
# A cross-compiler, from a 3 state dynamical system written in MAX/MIN/NOT
# formulae to its canonical mod-3 polynomial representation or to a C-language
# simulator

import argparse
import os
import sys

from steady_cell_phenotype._util import get_text_resource
from steady_cell_phenotype.equation_system import *

####################################################################################################


def main():
    parser = argparse.ArgumentParser(
        description="Converter from MAX/MIN/NOT formulae to either \n"
        + "low-degree polynomials over F_3 or a C-language simulator"
    )
    # noinspection SpellCheckingInspection
    parser.add_argument(
        "-i",
        "--inputfile",
        help="input filename containing MAX/MIN/NOT formulae. required. ",
        type=str,
    )
    # noinspection SpellCheckingInspection
    parser.add_argument(
        "-o",
        "--outputfile",
        help="output filename for the polynomial formulae. "
        "if not provided, stdout is used",
        type=str,
    )
    parser.add_argument(
        "-n",
        "--non_descriptive",
        action="store_true",
        help="use non-descriptive names for variables",
    )
    parser.add_argument(
        "-no-polys",
        action="store_true",
        help="do not output polynomials, used by default when "
        "output is by simulator",
    )
    parser.add_argument(
        "-sim", action="store_true", help="output C-language simulator program"
    )
    parser.add_argument(
        "-graph", action="store_true", help="use the graph-creation simulator"
    )
    parser.add_argument(
        "-complete_search",
        action="store_true",
        help="completely search the state-space",
    )
    parser.add_argument(
        "-init-val",
        action="append",
        nargs="+",
        help="for simulators, fix initial values got some variables "
        "Ex: -init-val LIP 1",
    )
    parser.add_argument(
        "--count",
        type=int,
        help="number of random points tried by the simulator, default 1,000,000.\n"
        "Ignored if the -sim flag is not used",
    )
    parser.add_argument(
        "-c",
        "--continuous",
        action="store_true",
        help="generate polynomials for continuous system, applied before the "
        "self-power operation",
    )
    parser.add_argument(
        "-comit",
        "--continuous-omit",
        nargs="+",
        help="list of variables to _not_ apply continuity operation to",
    )
    parser.add_argument(
        "-power",
        "--self-power",
        type=int,
        help="gets polynomials for a power of the system. i.e. self-composition,"
        " power-1 times (default: 1) ignored for simulator"
        " Warning: This can take a long time!",
    )

    args = parser.parse_args()

    in_formulae: str = ""
    if args.inputfile:
        # read input formulae from file, first checking that the file is there, etc.
        try:
            if not os.path.isfile(args.inputfile):
                print("Input file does not exist")
            elif not os.access(args.inputfile, os.R_OK):
                print("Input file is not readable")
            with open(args.inputfile, "r") as in_file:
                in_formulae = "".join(in_file.readlines())
        except IOError:
            print("Error reading file")
    else:
        print("Input file required\n")
        parser.print_help()

    translate_symbol_names_to_xs = args.non_descriptive

    ################################################################################
    # parse the file into a system of equations and ensure consistency

    equation_system = EquationSystem.from_text(lines=in_formulae)

    if not equation_system.consistent():
        print("Inconsistent system of equations!")
        sys.exit(-1)

    ################################################################################
    # parse initial values

    initial_value_list = []
    # these may come as several lists
    if args.init_val is not None:
        for sublist in args.init_val:
            initial_value_list += sublist

    # sanity check
    if len(initial_value_list) % 2 != 0:
        print("parse error on initial values, count of parameters")
        sys.exit(-1)

    initial_value_set_variables = initial_value_list[::2]
    for variable_name in initial_value_set_variables:
        if variable_name not in equation_system.symbol_table():
            print("initial value set for variable not in system")
            sys.exit(-1)

    try:
        set_variables = initial_value_list[::2]
        initial_values = [int(val) % 3 for val in initial_value_list[1::2]]
    except ValueError:
        print("parse error on initial values, value is not an int")
        sys.exit(-1)

    initial_values = dict(zip(set_variables, initial_values))

    ################################################################################
    # impose continuity, if desired
    # TODO: non-polynomial continuity for the simulator?

    if not args.continuous and args.continuous_omit:
        print(
            "Asked to omit continuity for system which was not continuous."
            " This is probably an error; exiting."
        )
        sys.exit(-1)
    if args.continuous is not None and args.continuous:
        all_variables = [str(var) for var in equation_system.symbol_table()]
        if args.continuous_omit is not None:
            continuous_variables = [
                var for var in all_variables if var not in args.continuous_omit
            ]
        else:
            continuous_variables = all_variables
        equation_system = equation_system.continuous_polynomial_system(
            continuous_vars=continuous_variables
        )

    if args.sim or args.graph or args.complete_search:
        #################################################################
        # run the simulator
        #################################################################
        count = 1_000_000
        if args.count is not None:
            count = args.count

        if not args.no_polys:
            equation_system = equation_system.as_poly_system()

        use_graph = args.graph is not None and args.graph
        do_complete_search = args.complete_search is not None and args.complete_search

        output_as_program(
            equation_system=equation_system,
            output_file=args.outputfile,
            num_runs=count,
            graph=use_graph,
            complete_search=do_complete_search,
            initial_values=initial_values,
        )
    else:
        #################################################################
        # output polynomials
        #################################################################

        # do composition if requested
        if args.self_power is not None:
            if args.self_power <= 0:
                raise Exception("Can only take positive power composition of system!")
            else:
                # Tried to see if this would help, no difference for the system I tried
                # equation_system = equation_system.as_poly_system()
                equation_system = equation_system.self_compose(count=args.self_power)

        output_as_formulae(
            equation_system=equation_system,
            translate_symbol_names_to_xs=translate_symbol_names_to_xs,
            as_polynomials=not args.no_polys,
            output_file=args.outputfile,
        )


####################################################################################################


def output_as_formulae(
    equation_system,
    translate_symbol_names_to_xs=True,
    as_polynomials=True,
    output_file=None,
):
    if as_polynomials:
        out_str_formulae = equation_system.polynomial_output(
            translate_symbol_names=translate_symbol_names_to_xs
        )
    else:
        out_str_formulae = str(equation_system)

    if output_file:
        try:
            with open(output_file, "w") as file_out:
                file_out.write(out_str_formulae)
                file_out.write("\n")
        except IOError:
            print("Error writing to file")
    else:
        print(out_str_formulae)
        print()


####################################################################################################


def output_as_program(
    equation_system,
    output_file=None,
    num_runs=1_000_000,
    graph=False,
    complete_search=False,
    initial_values=None,
):
    # get symbols, in order
    symbols = tuple(equation_system.target_variables())

    # parameter lists
    typed_param_list = "int " + ", int ".join(symbols)
    param_list = ", ".join(symbols)

    # create the functions which compute the transitions
    output_functions = []
    function_names = []

    for target, formula in equation_system:
        if isinstance(formula, int):
            string_c_formula = str(formula)
        else:
            string_c_formula = formula.as_c_expression()

        function_name = target + "_next"
        function_names.append(function_name)

        function_template = """
int {function_name}({typed_param_list})
{{
  return ( {c_formula} ) % 3;
}}"""
        function = function_template.format(
            target=target,
            function_name=function_name,
            typed_param_list=typed_param_list,
            c_formula=string_c_formula,
        )

        output_functions.append(function)

    # declarations of state variables
    variable_declaration = "\n".join(
        [
            "    int {name}_temp, {name};".format(name=symbol)
            for symbol, formula in equation_system
        ]
    )

    # random state initializer
    variable_initialization = "\n".join(
        [
            "    {name} = {value} % 3;".format(name=symbol, value=int(formula))
            if isinstance(formula, int) or formula.is_constant()
            else "    {name} = {value} % 3;".format(
                name=symbol, value=initial_values[symbol]
            )
            if symbol in initial_values
            else "    {name} = random() % 3;".format(name=symbol)
            for symbol, formula in equation_system
        ]
    )

    # deterministic search for loops
    state_for_loops_head = " ".join(
        [
            " int {name}_init_search = {value} % 3; ".format(
                name=symbol, value=int(formula)
            )
            if isinstance(formula, int) or formula.is_constant()
            else "  int {name}_init_search = {value} % 3;".format(
                name=symbol, value=initial_values[symbol]
            )
            if symbol in initial_values
            else "for(int {name}_init_search = 0;"
            " {name}_init_search < 3;"
            " {name}_init_search++ ) {{ ".format(name=symbol)
            for symbol, formula in equation_system
        ]
    )

    state_for_loops_tail = " ".join(
        [
            " "
            if isinstance(formula, int)
            or formula.is_constant()
            or symbol in initial_values
            else "}"
            for symbol, formula in equation_system
        ]
    )

    fixed_variable_initialization = "\n".join(
        [
            "    {name} = {name}_init_search;".format(name=symbol)
            for symbol, formula in equation_system
        ]
    )

    # run update, saving to temp variables
    def update_to_temp(indent=6):
        update_template = (indent * " ") + "{symbol}_temp = {func}({params});"
        return "\n".join(
            [
                update_template.format(symbol=symbol, func=fn_name, params=param_list)
                for symbol, fn_name in zip(symbols, function_names)
            ]
        )

    # copy the temp variables over
    def copy_temp_vars(indent=6):
        copy_template = (indent * " ") + "{symbol} = {symbol}_temp;"
        return "\n".join([copy_template.format(symbol=symbol) for symbol in symbols])

    # stash current "initial" state for a cycle
    def state_stash(indent=4):
        stash_template = (indent * " ") + "int {symbol}_init = {symbol};"
        return "\n".join([stash_template.format(symbol=symbol) for symbol in symbols])

    # check against "initial" state
    # noinspection PyUnusedLocal
    def neq_check(indent=4):
        neq_check_template = "{symbol} == {symbol}_init"
        return (
            "!("
            + " && ".join(
                [neq_check_template.format(symbol=symbol) for symbol in symbols]
            )
            + ")"
        )

    # print it out
    def print_state(indent=12):
        print_template = (indent * " ") + 'printf("\\"{symbol}\\":%u, ", {symbol});'
        print_template_last = (indent * " ") + 'printf("\\"{symbol}\\":%u ", {symbol});'
        header = (indent * " ") + 'printf("{ ");\n'
        footer = (indent * " ") + 'printf("}");\n'
        return (
            header
            + "\n".join(
                [
                    print_template.format(symbol=symbol)
                    if num != len(symbols) - 1
                    else print_template_last.format(symbol=symbol)
                    for num, symbol in enumerate(symbols)
                ]
            )
            + "\n"
            + footer
        )

    # hash functions
    hash_one = "\n".join(
        "  accumulator = 3*accumulator + {symbol};".format(symbol=symbol)
        for symbol in symbols
    )
    hash_two = "\n".join(
        "  accumulator = 5*accumulator + {symbol};".format(symbol=symbol)
        for symbol in symbols
    )
    hash_three = "\n".join(
        "  accumulator = 7*accumulator + {symbol};".format(symbol=symbol)
        for symbol in symbols
    )

    if graph:
        # alternate for graph!
        template = get_text_resource("gv-model-template.c")
    elif complete_search:
        template = get_text_resource("complete-search-model-template.c")
    else:
        # load the "big" template
        template = get_text_resource("model-template.c")

    # and use it
    program_text = template.format(
        param_list=param_list,
        typed_param_list=typed_param_list,
        accumulate_hash_one=hash_one,
        accumulate_hash_two=hash_two,
        accumulate_hash_three=hash_three,
        update_functions="\n".join(output_functions),
        compute_next6=update_to_temp(indent=6),
        compute_next8=update_to_temp(indent=8),
        compute_next12=update_to_temp(indent=12),
        copy6=copy_temp_vars(indent=6),
        copy8=copy_temp_vars(indent=8),
        copy12=copy_temp_vars(indent=12),
        declare_variables=variable_declaration,
        initialize_variables=variable_initialization,
        print_state=print_state(),
        variable_stash=state_stash(indent=4),
        neq_check=neq_check(indent=4),
        num_runs=num_runs,
        state_for_loops_head=state_for_loops_head,
        state_for_loops_tail=state_for_loops_tail,
        fixed_variable_initialization=fixed_variable_initialization,
    )

    if output_file:
        try:
            with open(output_file, "w") as out_file:
                out_file.write(program_text)
        except IOError:
            print("Error writing to file")
    else:
        print(program_text)


if __name__ == "__main__":
    # execute only if run as a script
    main()



import functools
import itertools
import math
import tempfile
from collections import defaultdict
from math import ceil
from typing import Dict, Iterator, List, NamedTuple, Tuple

import matplotlib
import matplotlib.pyplot as plt
import numba
import numpy as np
import pathos
from flask import Markup, make_response, render_template

from steady_cell_phenotype._util import (BinCounter, HashableNdArray, batcher,
                                         complete_search_generator,
                                         get_phased_trajectory,
                                         process_model_text,
                                         random_search_generator)

matplotlib.use("agg")


class TrajectoryStatistics(NamedTuple):
    trajectory_length_counts: Dict[HashableNdArray, BinCounter]
    data_counts: Dict[HashableNdArray, np.ndarray]
    means: Dict[HashableNdArray, np.ndarray]
    variances: Dict[HashableNdArray, np.ndarray]

    @classmethod
    def make(cls, num_variables: int):
        return TrajectoryStatistics(
            trajectory_length_counts=defaultdict(lambda: BinCounter()),
            data_counts=defaultdict(lambda: np.zeros(shape=0, dtype=int)),
            means=defaultdict(lambda: np.zeros(shape=(0, num_variables))),
            variances=defaultdict(lambda: np.zeros(shape=(0, num_variables))),
        )


def batch_trajectory_process(batch, update_fn) -> TrajectoryStatistics:
    num_samples = batch.shape[0]
    num_variables = batch.shape[1]

    trajectory_statistics = TrajectoryStatistics.make(num_variables)
    # unpack
    trajectory_length_counts = trajectory_statistics.trajectory_length_counts
    data_counts = trajectory_statistics.data_counts
    means = trajectory_statistics.means
    scaled_variances = trajectory_statistics.variances

    for batch_idx in range(num_samples):
        trajectory, phase_state = get_phased_trajectory(batch[batch_idx, :], update_fn)
        trajectory_length_counts[phase_state].add(len(trajectory))

        # extract arrays, to reduce dict lookup again
        data_count = data_counts[phase_state]
        mean = means[phase_state]
        scaled_variance = scaled_variances[phase_state]

        old_len: int = data_count.shape[0]
        trajectory_len: int = trajectory.shape[0]

        if trajectory_len == 0:
            continue

        # resize, if necessary
        if old_len < trajectory_len:
            # extend data_count
            new_data_count: np.ndarray = np.zeros(shape=trajectory_len, dtype=int)
            new_data_count[:old_len] = data_count
            data_count: np.ndarray = new_data_count
            data_counts[phase_state] = new_data_count
            # extend means
            new_mean: np.ndarray = np.zeros(shape=(trajectory_len, num_variables))
            new_mean[:old_len, :] = mean
            mean = new_mean
            means[phase_state] = new_mean
            # extend scaled variance
            new_scaled_variance: np.ndarray = np.zeros(
                shape=(trajectory_len, num_variables)
            )
            new_scaled_variance[:old_len] = scaled_variance
            scaled_variance = new_scaled_variance
            scaled_variances[phase_state] = new_scaled_variance

        # do it once
        reversed_trajectory = trajectory[::-1, :]

        # Welford's online algorithm for mean and variance calculation. See Knuth Vol 2, pg 232
        data_count[:trajectory_len] += 1
        old_mean = np.array(mean[:trajectory_len])  # copy
        mean[:trajectory_len] += (reversed_trajectory - old_mean) / np.expand_dims(
            data_count[:trajectory_len], axis=-1
        )
        scaled_variance[:trajectory_len] += (reversed_trajectory - old_mean) * (
            reversed_trajectory - mean[:trajectory_len]
        )

    # need to create a new tuple, as means and scaled variances can get clobbered.
    return TrajectoryStatistics(
        trajectory_length_counts=trajectory_length_counts,
        data_counts=data_counts,
        means=means,
        variances=scaled_variances,
    )


def reducer(
    init_stats: TrajectoryStatistics, new_stats: TrajectoryStatistics
) -> TrajectoryStatistics:
    # unpack
    trajectory_length_counts: Dict[HashableNdArray, BinCounter]
    data_counts: Dict[HashableNdArray, np.ndarray]
    means: Dict[HashableNdArray, np.ndarray]
    variances: Dict[HashableNdArray, np.ndarray]
    trajectory_length_counts, data_counts, means, variances = init_stats

    # unpack
    next_trajectory_length_counts: Dict[HashableNdArray, BinCounter]
    next_data_counts: Dict[HashableNdArray, np.ndarray]
    next_means: Dict[HashableNdArray, np.ndarray]
    next_scaled_variances: Dict[HashableNdArray, np.ndarray]
    (
        next_trajectory_length_counts,
        next_data_counts,
        next_means,
        next_scaled_variances,
    ) = new_stats

    for phase_point in next_trajectory_length_counts.keys():
        num_variables = next_means[phase_point].shape[1]
        if phase_point not in trajectory_length_counts:
            # just copy it
            trajectory_length_counts[phase_point] = next_trajectory_length_counts[
                phase_point
            ]
            data_counts[phase_point] = next_data_counts[phase_point]
            means[phase_point] = next_means[phase_point]
            # remember to scale!
            variances[phase_point] = next_scaled_variances[
                phase_point
            ] / np.expand_dims(data_counts[phase_point], axis=-1)
        else:
            # merge
            trajectory_length_counts[phase_point].add(
                next_trajectory_length_counts[phase_point]
            )
            new_max_len = trajectory_length_counts[phase_point].max

            # data counts
            old_data_count: np.ndarray = data_counts[phase_point]
            next_data_count: np.ndarray = next_data_counts[phase_point]
            new_data_counts = np.zeros(shape=new_max_len, dtype=int)
            new_data_counts[: old_data_count.shape[0]] = old_data_count
            new_data_counts[: next_data_count.shape[0]] += next_data_count

            data_counts[phase_point] = new_data_counts

            # means
            old_mean: np.ndarray = means[phase_point]
            next_mean: np.ndarray = next_means[phase_point]
            new_mean: np.ndarray = np.zeros(
                shape=(new_max_len, num_variables), dtype=np.float64
            )
            max_overlap = min(old_mean.shape[0], next_mean.shape[0])
            # weighted average on the overlap, copy over on the tail
            new_mean[:max_overlap] = (
                np.expand_dims(old_data_count[:max_overlap], axis=-1)
                * old_mean[:max_overlap]
                + np.expand_dims(next_data_count[:max_overlap], axis=-1)
                * next_mean[:max_overlap]
            ) / np.expand_dims(new_data_counts[:max_overlap], axis=-1)
            if old_mean.shape[0] > next_mean.shape[0]:
                new_mean[max_overlap:, :] = old_mean[max_overlap:, :]
            else:
                new_mean[max_overlap:, :] = next_mean[max_overlap:, :]

            means[phase_point] = new_mean

            # scaled variances
            old_variance: np.ndarray = variances[phase_point]
            next_scaled_variance: np.ndarray = next_scaled_variances[phase_point]
            new_variance: np.ndarray = np.zeros(
                shape=(new_max_len, num_variables), dtype=np.float64
            )
            # "weighted average" on the overlap, copy over on the tail
            old_proportion = np.expand_dims(
                old_data_count[:max_overlap] / new_data_counts[:max_overlap], axis=-1
            )
            next_proportion = np.expand_dims(
                next_data_count[:max_overlap] / new_data_counts[:max_overlap], axis=-1
            )
            new_variance[:max_overlap] = (
                old_proportion * old_variance[:max_overlap]
                + next_proportion
                * (
                    next_scaled_variance[:max_overlap]
                    / np.expand_dims(next_data_count[:max_overlap], axis=-1)
                )
                + ((old_mean[:max_overlap] - next_mean[:max_overlap]) ** 2)
                * old_proportion
                * next_proportion
            )
            if old_variance.shape[0] > next_scaled_variance.shape[0]:
                new_variance[max_overlap:, :] = old_variance[max_overlap:, :]
            else:
                new_variance[max_overlap:, :] = next_scaled_variance[
                    max_overlap:, :
                ] / np.expand_dims(next_data_count[max_overlap:], axis=-1)

            variances[phase_point] = new_variance

    return TrajectoryStatistics(trajectory_length_counts, data_counts, means, variances)


def compute_cycles(
    *,
    model_text: str,
    knockouts: Dict[str, int],
    continuous: Dict[str, bool],
    num_iterations: int,
    visualize_variables: Dict[str, bool],
):
    """

    Parameters
    ----------
    model_text
    knockouts
    continuous
    num_iterations
    visualize_variables

    Returns
    -------

    """
    # turn boolean-valued dictionary to list of variables to visualize
    visualized_variables: Tuple[str] = tuple(
        var for var in visualize_variables if visualize_variables[var]
    )

    # create an update function and equation system
    variables, update_fn, equation_system = process_model_text(
        model_text, knockouts, continuous
    )

    # compile and warm up the jitter
    update_fn = numba.jit(update_fn)
    update_fn(np.zeros(len(variables), dtype=np.int64))

    # associate variable names with their index in vectors
    variable_idx: Dict[str, int] = dict(zip(variables, range(len(variables))))

    constants: List[str] = equation_system.constant_variables()
    constants_vals: Dict[str, int] = {
        const: int(equation_system[const]) for const in constants
    }

    # decide if we will perform a complete state space search or not
    state_space_size = 3 ** (len(variables) - len(constants))
    perform_complete_search = state_space_size <= num_iterations

    state_generator: Iterator[np.ndarray]
    if perform_complete_search:
        state_generator = complete_search_generator(
            variables=variables, constants_vals=constants_vals
        )
        num_iterations = state_space_size
    else:
        state_generator = random_search_generator(
            num_iterations=num_iterations,
            variables=variables,
            constants_vals=constants_vals,
        )

    max_threads = max(
        1,
        min(
            pathos.multiprocessing.cpu_count() - 1, math.floor(state_space_size / 1000)
        ),
    )

    batch_generator = batcher(
        state_generator, variables, batch_size=min(1000, num_iterations // max_threads)
    )

    trajectory_length_counts: Dict[HashableNdArray, BinCounter] = dict()
    data_counts: Dict[HashableNdArray, np.ndarray] = dict()
    means: Dict[HashableNdArray, np.ndarray] = dict()
    variances: Dict[HashableNdArray, np.ndarray] = dict()
    with pathos.multiprocessing.ProcessPool(nodes=max_threads) as pool:
        trajectory_length_counts, data_counts, means, variances = functools.reduce(
            reducer,
            pool.uimap(
                batch_trajectory_process, batch_generator, itertools.repeat(update_fn)
            ),
            TrajectoryStatistics(
                trajectory_length_counts, data_counts, means, variances
            ),
        )

    phase_points = trajectory_length_counts.keys()

    # recreate limit cycles from phase point keys
    limit_cycles: Dict[
        HashableNdArray, List
    ] = dict()  # record limit sets with their ordering. i.e. as cycles
    for phase_point in phase_points:
        cycle = list()
        t_state = phase_point
        state = phase_point.array
        # do-while
        cycle.append(t_state)
        state = update_fn(state)
        t_state = HashableNdArray(state)
        while t_state != phase_point:
            cycle.append(t_state)
            state = update_fn(state)
            t_state = HashableNdArray(state)
        # record cycle
        limit_cycles[phase_point] = cycle

    # create images for each variable
    limit_set_stats_images = {phase_point: dict() for phase_point in phase_points}
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        plt.rcParams["svg.fonttype"] = "none"
        for phase_point, var in itertools.product(phase_points, visualized_variables):
            cycle = limit_cycles[phase_point]
            var_idx = variable_idx[var]

            var_means = np.flipud(means[phase_point][:, var_idx])
            var_stdevs = np.flipud(np.sqrt(variances[phase_point][:, var_idx]))

            plt.figure(figsize=(6, 4))

            # means, since phased will fix on single value at end
            plt.plot(list(var_means) + [cycle[0][var_idx]], color="#1f3d87")
            plt.fill_between(
                range(len(var_means) + 1),
                list(var_means - var_stdevs) + [cycle[0][var_idx]],
                list(var_means + var_stdevs) + [cycle[0][var_idx]],
                color="grey",
                alpha=0.25,
            )
            # dashed cycle in `converging` region
            plt.plot(
                range(len(var_means) - len(cycle), len(var_means) + 1),
                [cycle[idx][var_idx] for idx in range(len(cycle))]
                + [cycle[0][var_idx]],
                color="#F24F00",
                linestyle="dashed",
            )
            # solid cycle in `converged` region
            plt.plot(
                range(len(var_means), len(var_means) + len(cycle) + 1),
                [cycle[idx][var_idx] for idx in range(len(cycle))]
                + [cycle[0][var_idx]],
                color="#F24F00",
            )

            plt.title(f"Phased Envelope of trajectories for {var}")

            plt.xlabel("Distance to phase-fixing point on limit cycle")
            max_tick_goal = 40
            tick_gap = max(1, ceil((len(var_means) + len(cycle) + 1) / max_tick_goal))
            x_tick_locations = [
                x
                for x in range(0, len(var_means) + len(cycle) + 1)
                if (x - len(var_means)) % tick_gap == 0
            ]
            x_tick_labels = list(map(lambda x: len(var_means) - x, x_tick_locations))
            plt.xticks(x_tick_locations, x_tick_labels, rotation=90)
            plt.xlim([0, len(var_means) + len(cycle)])
            plt.axvline(len(var_means), linestyle="dotted", color="grey")
            plt.axvline(len(var_means) - len(cycle), linestyle="dotted", color="grey")

            plt.ylabel("State")
            plt.yticks([0, 1, 2])
            for y_val in range(3):
                plt.axhline(y=y_val, linestyle="dotted", color="grey", alpha=0.5)
            plt.ylim([0 - 0.1, 2 + 0.1])

            plt.tight_layout()

            image_filename = "dist-" + str(hash(phase_point)) + "-" + var + ".svg"
            plt.savefig(
                tmp_dir_name + "/" + image_filename, transparent=True, pad_inches=0.0
            )
            plt.close()
            with open(tmp_dir_name + "/" + image_filename, "r") as image:
                limit_set_stats_images[phase_point][var] = Markup(image.read())

    # print('var images complete')

    # cycle list sorted by frequency
    cycle_list = list(limit_cycles.values())
    cycle_list.sort(
        key=lambda cyc: trajectory_length_counts[cyc[0]].total(), reverse=True
    )

    # tally of number of cycles of each length, sorted by cycle length
    cycle_len_counts = defaultdict(lambda: 0)
    for cycle in cycle_list:
        cycle_len_counts[len(cycle)] += 1
    cycle_len_counts = tuple(sorted(cycle_len_counts.items()))

    length_distribution_images = dict()
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        # create length distribution plots
        plt.rcParams["svg.fonttype"] = "none"
        for phase_point in phase_points:
            length_dist = trajectory_length_counts[phase_point].trimmed_bins()
            plt.figure(figsize=(4, 3))
            plt.bar(x=range(len(length_dist)), height=length_dist, color="#002868")
            plt.title("Distribution of lengths of paths")
            plt.xlabel("Length of path")
            plt.ylabel("Number of states")
            plt.tight_layout()
            image_filename = "dist-" + str(hash(phase_point)) + ".svg"
            plt.savefig(
                tmp_dir_name + "/" + image_filename, transparent=True, pad_inches=0.0
            )
            plt.close()
            with open(tmp_dir_name + "/" + image_filename, "r") as image:
                length_distribution_images[phase_point] = Markup(image.read())

    def get_key(cyc):
        return cyc[0]

    def get_count(cyc):
        return trajectory_length_counts[get_key(cyc)].total()

    cycles = [
        {
            "states": cycle,
            "len": len(cycle),
            "count": get_count(cycle),
            "percent": 100 * get_count(cycle) / num_iterations,
            "len-dist-image": length_distribution_images[get_key(cycle)]
            if get_key(cycle) in length_distribution_images
            else "",
            "limit-set-stats-images": limit_set_stats_images[get_key(cycle)]
            if get_key(cycle) in limit_set_stats_images
            else dict(),
        }
        for cycle in cycle_list
    ]

    # respond with the results-of-computation page
    return make_response(
        render_template(
            "compute-cycles.html",
            cycles=cycles,
            variables=variables,
            cycle_len_counts=cycle_len_counts,
            complete_results=perform_complete_search,
        )
    )



from __future__ import annotations

from copy import copy
from enum import Enum
from itertools import product
from typing import Dict, List, Set, Tuple, Union

import numpy as np
from attr import attrib, attrs
from bs4 import BeautifulSoup
from bs4.element import Tag

PRIME = 3
# When MATHML_SANITY is `True`, allows the use of the <rem/> function. When `False`, this function
# is emulated. To conform to the SBML specification, we have to set sanity to false.
# (Tell me how you _really_ feel...)
MATHML_SANITY = False

ExpressionOrInt = Union[int, "Expression"]


class Operation(Enum):
    PLUS = "PLUS"
    MINUS = "MINUS"
    TIMES = "TIMES"
    EXP = "EXP"
    MAX = "MAX"
    MIN = "MIN"
    CONT = "CONT"
    NOT = "NOT"


####################################################################################################


def h(x: int, fx: int):
    """helper function as in the PLoS article, doi:10.1371/journal.pcbi.1005352.t003 pg 16/24"""
    fx %= PRIME
    x %= PRIME
    if fx > x:
        return x + 1
    elif fx < x:
        return x - 1
    else:
        return x


####################################################################################################


def inner_mathml_constant(value: int) -> Tag:
    constant_tag = Tag(name="cn", is_xml=True, attrs={"type": "integer"})
    constant_tag.append(str(value))
    return constant_tag


def inner_mathml_variable(var_name: str) -> Tag:
    var_tag = Tag(name="ci", is_xml=True)
    var_tag.append(var_name)
    return var_tag


def function_inner_mathml(function_name: str, operands: List[Tag]) -> Tag:
    # Note: up to you to do any copying of the operands that is needed
    top_tag = Tag(name="apply", is_xml=True)
    top_tag.append(Tag(name=function_name, is_xml=True, can_be_empty_element=True))
    for operand in operands:
        top_tag.append(operand)
    return top_tag


def single_var_to_power_inner_mathml(var_name: str, power: int) -> Tag:
    """Handle var^power"""
    if power == 0:
        return inner_mathml_constant(1)
    elif power == 1:
        return inner_mathml_variable(var_name)
    else:
        return function_inner_mathml(
            "power", [inner_mathml_variable(var_name), inner_mathml_constant(power)]
        )


def wrap_with_modulus_inner_mathml(
    expression: Tag, base: int, sane: bool = False
) -> Tag:
    """
    Wrap some inner mathml with a modulus operation.

    SBML does not support the <rem/> ("remainder") tag, which is a pain. So we have a sanity flag,
    `sane`, turned off by default for compatibility, which turns the simpler expression on.

    Parameters
    ----------
    expression: Tag
        The expression to wrap.
    base: int
        The base for the modulus.
    sane: bool
        Set sane to False to conform to the SBML subset of MathML. True allows a broader range
        of functions. e.g. <rem/>, the remainder


    Returns
    -------
    Tag
        Wrapped tag
    """
    if sane:
        return function_inner_mathml("rem", [expression, inner_mathml_constant(base)])
    else:
        # a mod b = a - b * floor(a/b)
        return function_inner_mathml(
            "minus",
            [
                expression,
                function_inner_mathml(
                    "times",
                    [
                        inner_mathml_constant(base),
                        function_inner_mathml(
                            "floor",
                            [
                                function_inner_mathml(
                                    "divide",
                                    [copy(expression), inner_mathml_constant(base)],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )


####################################################################################################
# monomial and sparse polynomial classes. These should be faster than the sympy versions due to
# their reduced scope.
####################################################################################################


class Expression(object):
    def __add__(self, other):
        return BinaryOperation("PLUS", self, other)

    __radd__ = __add__

    def __sub__(self, other):
        return BinaryOperation("MINUS", self, other)

    def __mul__(self, other):
        return BinaryOperation("TIMES", self, other)

    __rmul__ = __mul__

    def __neg__(self):
        return UnaryRelation("MINUS", self)

    def __pow__(self, power, modulo=None):
        return BinaryOperation("EXP", self, power)

    # def __divmod__(self, other):
    #    raise NotImplementedError("division, modulus not implemented")

    # def __truediv__(self, other):
    #    raise NotImplementedError("truediv not implemented")

    # def __floordiv__(self, other):
    #    raise NotImplementedError("floordiv not implemented")

    def eval(self, variable_dict):
        """
        Evaluates the expression.

        Parameters
        ----------
        variable_dict
            variable_dict is expected to be a dict containing str:Expression or
            Monomial:Expression pairs. The latter are constrained to be of single-variable type.

        Returns
        -------

        """
        raise NotImplementedError("eval() unimplemented in " + str(type(self)))

    def is_constant(self):
        raise NotImplementedError("is_constant() unimplemented in " + str(type(self)))

    def as_c_expression(self):
        raise NotImplementedError(
            "as_c_expression() unimplemented in " + str(type(self))
        )

    def as_polynomial(self) -> Union[int, Expression]:
        raise NotImplementedError("as_polynomial() unimplemented in " + str(type(self)))

    # def as_sympy(self):
    #     """
    #     converts to sympy expression
    #
    #     Returns
    #     -------
    #     sympy expression
    #     """
    #     raise NotImplementedError("as_sympy() unimplemented in " + str(type(self)))

    def as_numpy_str(self, variables: Tuple[str]) -> str:
        """
        Returns numpy-based function of variables.

        Order corresponds to that given in the variables parameter

        Parameters
        ----------
        variables

        Returns
        -------
        lambda with len(variables) parameters
        """
        raise NotImplementedError("as_numpy_str() unimplemented in " + str(type(self)))

    def as_mathml(self) -> BeautifulSoup:
        """
        Finds a mathml-based representation of the expression.

        Returns
        -------
        BeautifulSoup
        """
        soup = BeautifulSoup(features="xml")
        mathml = soup.new_tag(
            "math", attrs={"xmlns": "http://www.w3.org/1998/Math/MathML"}
        )
        soup.append(mathml)
        mathml.append(
            wrap_with_modulus_inner_mathml(
                self._make_inner_mathml(), base=PRIME, sane=MATHML_SANITY
            )
        )

        return soup

    def as_sbml_qual_relation(self, level: int) -> BeautifulSoup:
        """
        Finds a mathml represenation of 'level=expression', suitable for inclusion in SBML-qual.

        Parameters
        ----------
        level: int

        Returns
        -------
        BeautifulSoup
        """

        apply_tag = Tag(name="apply", is_xml=True)
        apply_tag.append(Tag(name="eq", is_xml=True, can_be_empty_element=True))
        apply_tag.append(inner_mathml_constant(level))
        apply_tag.append(
            wrap_with_modulus_inner_mathml(
                self._make_inner_mathml(), base=PRIME, sane=MATHML_SANITY
            )
        )

        soup = BeautifulSoup(features="xml")
        mathml = soup.new_tag(
            "math", attrs={"xmlns": "http://www.w3.org/1998/Math/MathML"}
        )
        soup.append(mathml)
        mathml.append(apply_tag)

        return soup

    def _make_inner_mathml(self) -> Union[BeautifulSoup, Tag]:
        raise NotImplementedError("_inner_mathml() unimplemented in " + str(type(self)))

    def get_variable_set(self) -> Set[str]:
        """returns a set containing all variable which occur in this expression"""
        raise NotImplementedError("get_var_set() unimplemented in " + str(type(self)))

    def num_variables(self):
        """returns the number of variables which occur in this expression"""
        return len(self.get_variable_set())

    def rename_variables(self, name_dict: Dict[str, str]):
        """rename variables"""
        raise NotImplementedError(
            "rename_variables() unimplemented in " + str(type(self))
        )

    def continuous_function_version(self, control_variable):
        """
        Wrap this equation with the 'continuity controller' i.e. return CONT(control_variable,self)
        :param control_variable: variable or string
        :return: functional continuous version
        """
        if self.is_constant():
            return self

        if isinstance(control_variable, str):
            control_variable = Monomial.as_var(control_variable)

        return Function("CONT", [control_variable, self])

    ################################################################################################
    #
    # the following method converts a system of equations into one which is "continuous" in the
    # sense that application of the system does not change the per-coordinate values by more than 1.
    # This is accomplished by a type of curve fitting. Fortunately, the formula for this
    #
    # g(x) = sum_{c\in \F_3^n} h(c) prod_{j=0}^n (1-(x_j-c_j)**2)
    #
    # (as seen in the PLoS article, doi:10.1371/journal.pcbi.1005352.t003 pg 16/24) admits a
    # recursive formulation. That is, for a polynomial x_k = f_k(x_0,x_1,...,x_l) we can select one
    # of the variables, say x_0 and reduce the polynomial each of 3-ways x_0=0, x_0=1, and x_0=2.
    # This correspondingly divides the sum into those which have each of the 3 types of terms
    # (1-(x_0-c_0)**2) for c_0=0, c_0=1, and c_0=2
    #
    # fortunately, (1-(x_j-0)**2)+(1-(x_j-1)**2)+(1-(x_j-2)**2) = 1 so if the evaluations of f
    # become constant or even simply eliminate a variable, we need no longer consider that variable.
    #
    # recursion proceeds by eliminating variables in this manner, multiplying by the appropriate
    # fitting term (1-(x_j-c_j)**2) (c_j being the evaluated value of x_j) on the way up.
    #
    # this comment is not really the place for a full proof of this method, but the proof is easily
    # obtained from the above.
    #
    ################################################################################################

    def continuous_polynomial_version(self, control_variable):
        if self.is_constant():
            return self

        if isinstance(control_variable, str):
            control_variable = Monomial.as_var(control_variable)

        # as the control variable is special (due to use in the 'h' function),
        # we will need to go through the procedure for it separately, first
        accumulator = Mod3Poly.zero()
        for control_variable_value in range(3):
            evaluated_poly = self.eval({control_variable: control_variable_value})
            if is_integer(evaluated_poly) or evaluated_poly.is_constant():
                computed_value: int = int(evaluated_poly)
                continuous_value = h(control_variable_value, computed_value)
                accumulator += continuous_value * (
                    1 - (control_variable - control_variable_value) ** 2
                )
            else:
                accumulator += evaluated_poly.continuous_version_helper(
                    control_variable_value
                ) * (1 - (control_variable - control_variable_value) ** 2)
        return accumulator

    def continuous_version_helper(self, control_variable_value):
        # find some free variable
        free_variable = tuple(self.get_variable_set())[0]
        if isinstance(free_variable, str):
            free_variable = Monomial.as_var(free_variable)

        # iterate over the ways of setting that variable: 0, 1, 2
        accumulator = Mod3Poly.zero()
        for free_variable_value in range(3):
            evaluated_poly = self.eval({free_variable: free_variable_value})
            if is_integer(evaluated_poly) or evaluated_poly.is_constant():
                computed_value: int = int(evaluated_poly)
                continuous_value = h(control_variable_value, computed_value)
                accumulator += continuous_value * (
                    1 - (free_variable - free_variable_value) ** 2
                )
            else:
                accumulator += evaluated_poly.continuous_version_helper(
                    control_variable_value
                ) * (1 - (free_variable - free_variable_value) ** 2)
        return accumulator


####################################################################################################


def rename_helper(expression: Union[Expression, int], name_dict: Dict[str, str]):
    if is_integer(expression):
        return expression
    else:
        return expression.rename_variables(name_dict=name_dict)


####################################################################################################
# actions on expressions, suitable for conversion to polynomial form. Not best for simulator.


def mod_3(n):
    return n % 3


def not3(n):
    value = 2 + 2 * n
    if is_integer(value) or value.is_constant():
        return mod_3(int(value))
    else:
        return value


def max3(a, b):
    value = a + b + 2 * a * b + (a ** 2) * b + a * (b ** 2) + (a ** 2) * (b ** 2)
    if is_integer(value) or value.is_constant():
        return mod_3(int(value))
    else:
        return value


def min3(a, b):
    value = a * b + 2 * (a ** 2) * b + 2 * a * (b ** 2) + 2 * (a ** 2) * (b ** 2)
    if is_integer(value) or value.is_constant():
        return mod_3(int(value))
    else:
        return value


def is_integer(x):
    return isinstance(x, int) or isinstance(x, np.integer)


####################################################################################################


class Function(Expression):
    def __init__(self, function_name: str, expression_list):
        self._function_name = function_name
        self._expression_list: List[ExpressionOrInt] = expression_list

    def rename_variables(self, name_dict: Dict[str, str]):
        renamed_parameters = [
            rename_helper(expr, name_dict) for expr in self._expression_list
        ]
        return Function(self._function_name, renamed_parameters)

    def eval(self, variable_dict):
        # evaluate function parameters
        evaluated_expressions = [
            expr if is_integer(expr) else expr.eval(variable_dict)
            for expr in self._expression_list
        ]
        # simplify constants to ints, if possible
        evaluated_expressions = [
            int(expr) if is_integer(expr) or expr.is_constant() else expr
            for expr in evaluated_expressions
        ]

        if self._function_name == "MAX":
            assert len(evaluated_expressions) == 2, "wrong number of arguments for MAX"
            expr_one, expr_two = evaluated_expressions
            # if it can be computed directly, do it. otherwise, return in function form
            if is_integer(expr_one) and is_integer(expr_two):
                expr_one = mod_3(expr_one)
                expr_two = mod_3(expr_two)
                return max(expr_one, expr_two)
            elif is_integer(expr_one) and expr_one == 2:
                return 2
            elif is_integer(expr_one) and expr_one == 0:
                return expr_two
            elif is_integer(expr_two) and expr_two == 2:
                return 2
            elif is_integer(expr_two) and expr_two == 0:
                return expr_one
            else:
                return Function("MAX", [expr_one, expr_two])
        elif self._function_name == "MIN":
            assert len(evaluated_expressions) == 2, "wrong number of arguments for MIN"
            expr_one, expr_two = evaluated_expressions
            # if it can be computed directly, do it. otherwise, return in function form
            if is_integer(expr_one) and is_integer(expr_two):
                expr_one = mod_3(expr_one)
                expr_two = mod_3(expr_two)
                return min(expr_one, expr_two)
            elif is_integer(expr_one) and expr_one == 2:
                return expr_two
            elif is_integer(expr_one) and expr_one == 0:
                return 0
            elif is_integer(expr_two) and expr_two == 2:
                return expr_one
            elif is_integer(expr_two) and expr_two == 0:
                return 0
            else:
                return Function("MIN", [expr_one, expr_two])
        elif self._function_name == "CONT":
            assert len(evaluated_expressions) == 2, "wrong number of arguments for CONT"
            ctrl_var, expr = evaluated_expressions
            if is_integer(ctrl_var):
                raise Exception("Unsupported; nonsense")
            return Function("CONT", [ctrl_var, expr])
        elif self._function_name == "NOT":
            assert len(evaluated_expressions) == 1, "wrong number of arguments for NOT"
            expr = evaluated_expressions[0]
            # if it can be computed directly, do it. otherwise, return in function form
            if is_integer(expr):
                return not3(int(expr))
            else:
                return Function("NOT", [expr])
        else:
            raise Exception("cannot evaluate unknown function " + self._function_name)

    def is_constant(self):
        return all(
            is_integer(expr) or expr.is_constant() for expr in self._expression_list
        )

    def __str__(self):
        return (
            self._function_name
            + "("
            + ",".join([str(exp) for exp in self._expression_list])
            + ")"
        )

    __repr__ = __str__

    def as_c_expression(self):
        c_exprs = [
            str(expr) if is_integer(expr) else expr.as_c_expression()
            for expr in self._expression_list
        ]

        if self._function_name == "MAX":
            func_name = "mod3max"
        elif self._function_name == "MIN":
            func_name = "mod3min"
        elif self._function_name == "CONT":
            func_name = "mod3continuity"
        elif self._function_name == "NOT":
            func_name = "mod3not"
        else:
            raise Exception("Unknown binary relation: " + self._function_name)

        return func_name + "(" + ",".join(c_exprs) + ")"

    def as_polynomial(self):
        expressions_as_polynomials = [
            mod_3(expr) if is_integer(expr) else expr.as_polynomial()
            for expr in self._expression_list
        ]

        if self._function_name == "MAX":
            assert (
                len(expressions_as_polynomials) == 2
            ), "wrong number of arguments for MAX"
            return max3(expressions_as_polynomials[0], expressions_as_polynomials[1])

        elif self._function_name == "MIN":
            assert (
                len(expressions_as_polynomials) == 2
            ), "wrong number of arguments for MIN"
            return min3(expressions_as_polynomials[0], expressions_as_polynomials[1])

        elif self._function_name == "CONT":
            assert (
                len(expressions_as_polynomials) == 2
            ), "wrong number of arguments for CONT"
            return expressions_as_polynomials[1].continuous_polynomial_version(
                expressions_as_polynomials[0]
            )

        elif self._function_name == "NOT":
            assert (
                len(expressions_as_polynomials) == 1
            ), "wrong number of arguments for NOT"
            return not3(expressions_as_polynomials[0])

        else:
            raise Exception(
                "cannot evaluate unknown function "
                + self._function_name
                + " as a polynomial"
            )

    # def as_sympy(self):
    #
    #     def cont_sympy(control, expr):
    #         return expr if is_integer(expr) \
    #             else expr.continuous_polynomial_version(control)
    #
    #     def not_sympy(expr):
    #         return 1 - expr
    #
    #     # tuples are param-count, function
    #     functions = {'MAX': (2, sympy.Max),
    #                  'MIN': (2, sympy.Min),
    #                  'CONT': (2, cont_sympy),
    #                  'NOT': (1, not_sympy)}
    #
    #     if self._function_name not in functions:
    #         raise Exception("cannot evaluate unknown function " + self._function_name +
    #                         " as a sympy expression")
    #
    #     if len(self._expression_list) != functions[self._function_name][0]:
    #         raise Exception(f"Wrong number of arguments for {self._function_name}")
    #
    #     function = functions[self._function_name][1]
    #
    #     sympy_expressions = [sympy.Mod(expr, 3) if is_integer(expr)
    #                          else sympy.Mod(expr.as_sympy(), 3)
    #                          for expr in self._expression_list]
    #     return function(*sympy_expressions)

    def as_numpy_str(self, variables) -> str:
        np_parameter_strings = [
            str(expr) if is_integer(expr) else expr.as_numpy_str(variables)
            for expr in self._expression_list
        ]

        continuous_str = "( {0}+np.sign(np.mod({1},3)-np.mod({0},3)) )"
        max_str = "np.maximum(np.mod({0},3),np.mod({1},3))"
        min_str = "np.minimum(np.mod({0},3),np.mod({1},3))"
        not_str = "(2-({0}))"

        # tuples are param-count, function
        function_strings = {
            "MAX": (2, max_str),
            "MIN": (2, min_str),
            "CONT": (2, continuous_str),
            "NOT": (1, not_str),
        }

        if self._function_name not in function_strings:
            raise Exception(
                "cannot evaluate unknown function "
                + self._function_name
                + " as a numpy function"
            )

        if len(self._expression_list) != function_strings[self._function_name][0]:
            raise Exception(f"Wrong number of arguments for {self._function_name}")

        function = function_strings[self._function_name][1]

        return function.format(*np_parameter_strings)

    def _make_inner_mathml(self) -> Tag:
        apply_tag = Tag(name="apply", is_xml=True)

        mathml_function_strings = {
            "MAX": Tag(name="max", is_xml=True, can_be_empty_element=True),
            "MIN": Tag(name="max", is_xml=True, can_be_empty_element=True),
        }

        if self._function_name in mathml_function_strings:
            apply_tag.append(mathml_function_strings[self._function_name])
            for expression in self._expression_list:
                apply_tag.append(
                    wrap_with_modulus_inner_mathml(
                        expression._make_inner_mathml(), base=PRIME, sane=MATHML_SANITY
                    )
                )
        elif self._function_name == "NOT":
            # rewrite as (prime-1)-expression
            apply_tag.append(Tag(name="minus", is_xml=True, can_be_empty_element=True))
            apply_tag.append(inner_mathml_constant(PRIME - 1))
            apply_tag.append(
                wrap_with_modulus_inner_mathml(
                    self._expression_list[0]._make_inner_mathml(),
                    base=PRIME,
                    sane=MATHML_SANITY,
                )
            )
        else:
            raise NotImplementedError(
                "_inner_mathml() unimplemented in " + str(type(self))
            )

        return apply_tag

    def get_variable_set(self) -> Set[str]:
        var_set = set()
        for expr in self._expression_list:
            if not is_integer(expr):
                var_set = var_set.union(expr.get_variable_set())
        return var_set


class BinaryOperation(Expression):
    def __init__(
        self,
        relation_name,
        left_expression: ExpressionOrInt,
        right_expression: ExpressionOrInt,
    ):
        self._relation_name = relation_name
        self._left_expression: ExpressionOrInt = left_expression
        self._right_expression: ExpressionOrInt = right_expression

    def rename_variables(self, name_dict: Dict[str, str]):
        renamed_left_expression = rename_helper(self._left_expression, name_dict)
        renamed_right_expression = rename_helper(self._right_expression, name_dict)
        return BinaryOperation(
            self._relation_name,
            left_expression=renamed_left_expression,
            right_expression=renamed_right_expression,
        )

    def is_constant(self):
        return (
            is_integer(self._left_expression) or self._left_expression.is_constant()
        ) and (
            is_integer(self._right_expression) or self._right_expression.is_constant()
        )

    def eval(self, variable_dict) -> Union[Expression, int]:
        """
        Evaluate parameters, making them ints if possible.

        Parameters
        ----------
        variable_dict
            a dictionary of taking either single-term monomials or string (variable names) to ints

        Returns
        -------
        evaluated expression
        """
        evaled_left_expr = (
            self._left_expression
            if is_integer(self._left_expression)
            else self._left_expression.eval(variable_dict)
        )
        evaled_left_expr = (
            int(evaled_left_expr)
            if is_integer(evaled_left_expr) or evaled_left_expr.is_constant()
            else evaled_left_expr
        )

        evaled_right_expr = (
            self._right_expression
            if is_integer(self._right_expression)
            else self._right_expression.eval(variable_dict)
        )
        evaled_right_expr = (
            int(evaled_right_expr)
            if is_integer(evaled_right_expr) or evaled_right_expr.is_constant()
            else evaled_right_expr
        )

        operations = {
            "PLUS": lambda l, r: l + r,
            "MINUS": lambda l, r: l - r,
            "TIMES": lambda l, r: l * r,
            "EXP": lambda l, r: l ** r,
        }
        if self._relation_name in operations:
            return operations[self._relation_name](evaled_left_expr, evaled_right_expr)
        else:
            raise Exception(f"cannot evaluate unknown binary op: {self._relation_name}")

    def __str__(self):
        short_relation_names = {"PLUS": "+", "MINUS": "-", "TIMES": "*", "EXP": "^"}
        if self._relation_name in short_relation_names:
            short_relation_name = short_relation_names[self._relation_name]
        else:
            short_relation_name = "?"

        left_side = str(self._left_expression)
        if isinstance(self._left_expression, BinaryOperation):
            left_side = "(" + left_side + ")"

        right_side = str(self._right_expression)
        if isinstance(self._right_expression, BinaryOperation):
            right_side = "(" + right_side + ")"

        return left_side + short_relation_name + right_side

    __repr__ = __str__

    def as_c_expression(self):
        if is_integer(self._left_expression):
            left_c_expr = str(self._left_expression)
        else:
            left_c_expr = self._left_expression.as_c_expression()

        if is_integer(self._right_expression):
            right_c_expr = str(self._right_expression)
        else:
            right_c_expr = self._right_expression.as_c_expression()

        if self._relation_name == "PLUS":
            return "(" + left_c_expr + ")+(" + right_c_expr + ")"

        elif self._relation_name == "MINUS":
            return "(" + left_c_expr + ")-(" + right_c_expr + ")"

        elif self._relation_name == "TIMES":
            return "(" + left_c_expr + ")*(" + right_c_expr + ")"

        elif self._relation_name == "EXP":
            return "mod3pow(" + left_c_expr + "," + right_c_expr + ")"

        else:
            raise Exception("Unknown binary relation: " + self._relation_name)

    def as_polynomial(self):
        if is_integer(self._left_expression):
            left_poly = self._left_expression
        else:
            left_poly = self._left_expression.as_polynomial()

        if is_integer(self._right_expression):
            right_poly = self._right_expression
        else:
            right_poly = self._right_expression.as_polynomial()

        if self._relation_name == "PLUS":
            return left_poly + right_poly

        elif self._relation_name == "MINUS":
            return left_poly - right_poly

        elif self._relation_name == "TIMES":
            return left_poly * right_poly

        elif self._relation_name == "EXP":
            # simplify the exponent = 0, 1 cases
            if is_integer(right_poly):
                if right_poly == 0:
                    return 1
                elif right_poly == 1:
                    return left_poly
                else:
                    return left_poly ** right_poly
            else:
                return left_poly ** right_poly
        else:
            raise Exception("Unknown binary relation: " + self._relation_name)

    # def as_sympy(self):
    #     """
    #     Convert to sympy expression
    #     Returns
    #     -------
    #     sympy expression
    #     """
    #
    #     def simple_pow(left_exp, right_exp):
    #         # simplify the exponent = 0, 1 cases
    #         if is_integer(right_exp):
    #             if right_exp == 0:
    #                 return 1
    #             elif right_exp == 1:
    #                 return left_exp
    #             else:
    #                 return left_exp ** right_exp
    #         else:
    #             return left_exp ** right_exp
    #
    #     relations = {'PLUS': operator.add,
    #                  'MINUS': operator.sub,
    #                  'TIMES': operator.mul,
    #                  'EXP': simple_pow}
    #
    #     if self.relation_name not in relations:
    #         raise Exception("Unknown binary relation: " + self.relation_name)
    #
    #     lhs = (self._left_expression if is_integer(self._left_expression)
    #            else self._left_expression.as_sympy())
    #     rhs = (self._right_expression if is_integer(self._right_expression)
    #            else self._right_expression.as_sympy())
    #
    #     return relations[self.relation_name](lhs, rhs)

    def as_numpy_str(self, variables) -> str:
        """
        Convert to numpy function
        Parameters
        ----------
        variables

        Returns
        -------
        str version of numpy function
        """

        relations = {
            "PLUS": "(({0})+({1}))",
            "MINUS": "(({0})-({1}))",
            "TIMES": "(({0})*({1}))",
            "EXP": "(({0})**({1}))",
        }

        if self._relation_name not in relations:
            raise Exception("Unknown binary relation: " + self._relation_name)

        lhs = (
            str(self._left_expression)
            if is_integer(self._left_expression)
            else self._left_expression.as_numpy_str(variables)
        )
        rhs = (
            str(self._right_expression)
            if is_integer(self._right_expression)
            else self._right_expression.as_numpy_str(variables)
        )

        return relations[self._relation_name].format(lhs, rhs)

    def _make_inner_mathml(self) -> Tag:
        apply_tag = Tag(name="apply", is_xml=True)

        mathml_relations = {
            "PLUS": Tag(name="add", is_xml=True, can_be_empty_element=True),
            "MINUS": Tag(name="minus", is_xml=True, can_be_empty_element=True),
            "TIMES": Tag(name="times", is_xml=True, can_be_empty_element=True),
            "EXP": Tag(name="power", is_xml=True, can_be_empty_element=True),
        }

        if self._relation_name in mathml_relations:
            apply_tag.append(mathml_relations[self._relation_name])

            if (
                isinstance(self._left_expression, Expression)
                and not self._left_expression.is_constant()
            ):
                apply_tag.append(self._left_expression._make_inner_mathml())
            else:
                apply_tag.append(inner_mathml_constant(int(self._left_expression)))

            if (
                isinstance(self._right_expression, Expression)
                and not self._right_expression.is_constant()
            ):
                apply_tag.append(self._right_expression._make_inner_mathml())
            else:
                apply_tag.append(inner_mathml_constant(int(self._right_expression)))
        else:
            raise NotImplementedError(
                f"_inner_mathml() unimplemented for {self._relation_name} in {type(self)}"
            )

        return apply_tag

    def get_variable_set(self) -> Set[str]:
        var_set = set()
        if not is_integer(self._left_expression):
            var_set = var_set.union(self._left_expression.get_variable_set())
        if not is_integer(self._right_expression):
            var_set = var_set.union(self._right_expression.get_variable_set())
        return var_set


class UnaryRelation(Expression):
    def __init__(self, relation_name, expr):
        self._relation_name = relation_name
        self._expr: ExpressionOrInt = expr

    def rename_variables(self, name_dict: Dict[str, str]):
        return UnaryRelation(
            relation_name=self._relation_name, expr=rename_helper(self._expr, name_dict)
        )

    def is_constant(self):
        return self._expr.is_constant()

    def eval(self, variable_dict):
        if self._relation_name == "MINUS":
            if is_integer(self._expr):
                return (-1) * self._expr
            elif type(self._expr) == Expression:
                evaluated_subexpression = self._expr.eval(variable_dict)
                if (
                    is_integer(evaluated_subexpression)
                    or evaluated_subexpression.is_constant()
                ):
                    return (-1) * int(evaluated_subexpression)
                else:
                    return (-1) * evaluated_subexpression
        else:
            raise Exception(
                "UnaryRelation in bad state with unknown unary relation name"
            )

    def __str__(self) -> str:
        short_rel_name = str(self._relation_name)
        if self._relation_name == "MINUS":
            short_rel_name = "-"
        return short_rel_name + (
            "(" + str(self._expr) + ")"
            if type(self._expr) == BinaryOperation
            else str(self._expr)
        )

    __repr__ = __str__

    def as_c_expression(self):
        if is_integer(self._expr):
            c_exp = str(mod_3(self._expr))
        else:
            c_exp = self._expr.as_c_expression()

        if self._relation_name == "MINUS":
            return "-(" + c_exp + ")"
        else:
            raise Exception("Unknown binary relation: " + self._relation_name)

    def as_polynomial(self):
        if is_integer(self._expr) or self._expr.is_constant():
            poly = mod_3(int(self._expr))
        else:
            poly = self._expr.as_polynomial()

        if self._relation_name == "MINUS":
            return (-1) * poly
        else:
            raise Exception("Unknown unary relation: " + self._relation_name)

    # def as_sympy(self):
    #     """
    #     Convert to sympy expression
    #     Returns
    #     -------
    #     sympy expression
    #     """
    #
    #     relations = {"MINUS": operator.neg}
    #
    #     if self._relation_name not in relations:
    #         raise Exception("Unknown unary relation: " + self._relation_name)
    #
    #     expr = self._expr if is_integer(self._expr) else self._expr.as_sympy()
    #
    #     return relations[self._relation_name](expr)

    def as_numpy_str(self, variables):
        """
        Convert to numpy function
        Parameters
        ----------
        variables

        Returns
        -------
        str numpy-representation
        """

        relations = {"MINUS": "(-({0}))"}

        if self._relation_name not in relations:
            raise Exception("Unknown unary relation: " + self._relation_name)

        expr_str = (
            str(self._expr)
            if is_integer(self._expr)
            else self._expr.as_numpy_str(variables)
        )

        return relations[self._relation_name].format(expr_str)

    def _make_inner_mathml(self) -> Tag:
        apply_tag = Tag(name="apply", is_xml=True)

        mathml_relations = {
            "MINUS": Tag(name="minus", is_xml=True, can_be_empty_element=True),
        }

        if self._relation_name in mathml_relations:
            apply_tag.append(mathml_relations[self._relation_name])

            if isinstance(self._expr, Expression) and not self._expr.is_constant():
                apply_tag.append(self._expr._make_inner_mathml())
            else:
                apply_tag.append(inner_mathml_constant(int(self._expr)))
        else:
            raise NotImplementedError(
                f"_inner_mathml() unimplemented for {self._relation_name} in {type(self)}"
            )

        return apply_tag

    def get_variable_set(self) -> Set[str]:
        if is_integer(self._expr):
            return set()
        else:
            return self._expr.get_variable_set()


####################################################################################################


@attrs(init=False, cmp=False, repr=False, str=False)
class Monomial(Expression):
    """A class to encapsulate monomials reduced by x^3-x==0 for all variables x"""

    _power_dict: Dict[str, int] = attrib()

    def __init__(self, power_dict: Dict[str, int]):
        # copy over only those terms which actually appear
        self._power_dict = {
            str(var): power_dict[var] for var in power_dict if power_dict[var] != 0
        }
        for var in self._power_dict.keys():
            # while self._power_dict[var] < 0:
            #    self._power_dict[var] += 2     <--- replace with below
            assert self._power_dict[var] > 0  # b/c x^-1 isn't exactly x (i.e. when x=0)
            # while self._power_dict[var] >= 3:
            #    self._power_dict[var] -= 2     <--- replace with below
            self._power_dict[var] = 1 + ((-1 + self._power_dict[var]) % 2)

    def rename_variables(self, name_dict: Dict[str, str]):
        # this ends up a little more complicated than I was originally thinking, b/c
        # I would like to allow two variables to be updated to the same new name
        renamed_dict = dict()
        for variable, exponent in self._power_dict.items():
            name = variable
            if variable in name_dict:
                name = name_dict[variable]
            if name in renamed_dict:
                renamed_dict[name] += self._power_dict[variable]
                renamed_dict[name] = 1 + ((-1 + renamed_dict[name]) % 2)
            else:
                renamed_dict[name] = self._power_dict[variable]
        return Monomial(power_dict=renamed_dict)

    def as_polynomial(self):
        return self

    def is_constant(self):
        return len(self._power_dict) == 0

    def num_variables(self):
        return len(self._power_dict)

    def variable_list(self):
        return self._power_dict.keys()

    def eval(self, variable_dict: Dict) -> Expression:
        """
        Evaluates the monomial.

        Parameters
        ----------
        variable_dict
            variable_dict is expected to be a dict containing str:Expression or
            Monomial:Expression pairs. The latter are constrained to be of single-variable type.

        Returns
        -------
        Expression
        """
        if type(variable_dict) != dict:
            raise Exception("eval is not defined on this input")

        # sanitize inputs
        sanitized_variable_dict = dict()
        for variable, quantity in variable_dict.items():
            if type(variable) == str:
                sanitized_variable_dict.update({variable: variable_dict[variable]})
            elif type(variable) == Monomial:
                if variable.num_variables() != 1:
                    raise Exception(
                        "We do not know how to evaluate monomials of zero or"
                        " several variables to a single number"
                    )
                else:
                    variable_as_str = list(variable.variable_list())[0]
                    sanitized_variable_dict.update(
                        {variable_as_str: variable_dict[variable]}
                    )
        variable_dict = sanitized_variable_dict

        accumulator = Mod3Poly.one()
        for variable, quantity in self._power_dict.items():
            if variable in variable_dict.keys():
                accumulator *= variable_dict[variable] ** self._power_dict[variable]
            else:
                accumulator *= Monomial.as_var(variable) ** self._power_dict[variable]
        return accumulator

    def get_variable_set(self) -> Set[str]:
        """returns a set containing all variable which occur in this monomial"""
        return {var for var in self._power_dict if self._power_dict[var] != 0}

    @staticmethod
    def unit():
        """produces the unit, 1, as a monomial"""
        return Monomial(dict())

    @staticmethod
    def as_var(var_name: str):
        return Monomial({var_name: 1})

    def __mul__(self, other) -> Expression:
        if isinstance(other, Monomial):
            result_power_dict = self._power_dict.copy()
            for key in other._power_dict.keys():
                if key in result_power_dict.keys():
                    result_power_dict[key] += other._power_dict[key]
                    while result_power_dict[key] >= 3:
                        result_power_dict[key] -= 2
                else:
                    result_power_dict[key] = other._power_dict[key]
            return Monomial(result_power_dict)
        elif isinstance(other, Mod3Poly) or is_integer(other):
            return self.as_poly() * other
        else:
            return BinaryOperation("TIMES", self, other)

    __rmul__ = __mul__

    def __neg__(self):
        return (-1) * self

    def __pow__(self, power: ExpressionOrInt, **kwargs):
        if type(power) != int:
            if power.is_constant():
                power = int(power)
            else:
                raise NotImplementedError(
                    "Cannot raise an Expression to a non-integer power"
                )

        if power == 0:
            return Monomial.unit()
        elif power == 1:
            return self
        elif power == 2:
            return self * self
        # Now handle higher powers; probably not going to happen too much for this application

        # (int) half power root
        int_root = self ** (power // 2)
        if power % 2 == 0:
            return int_root * int_root
        else:
            return int_root * int_root * self

    def as_poly(self):
        """converts this monomial to a polynomial with only one term"""
        return Mod3Poly({self: 1})

    def __add__(self, other):
        if isinstance(other, Mod3Poly):
            return other + self.as_poly()
        elif isinstance(other, Monomial):
            return self.as_poly() + other.as_poly()
        elif is_integer(other):
            return self.as_poly() + other
        elif isinstance(other, Expression):
            return BinaryOperation("PLUS", self, other)
        else:
            raise TypeError(
                "unsupported operand type(s) for +: '{}' and '{}'".format(
                    self.__class__, type(other)
                )
            )

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return self + ((-1) * other)

    def __rsub__(self, other):
        return ((-1) * self) + other

    def __eq__(self, other):
        if type(other) == str:
            other = Monomial.as_var(other)
        if type(other) == Monomial:
            return self._power_dict == other._power_dict
        elif type(other) == Mod3Poly:
            if len(other.coeff_dict) == 1:
                monomial, coeff = list(other.coeff_dict)[0]
                return coeff == 1 and monomial == self
            else:
                return False
        elif is_integer(other) and self == Monomial.unit():
            return other == 1
        else:
            return False

    def __ne__(self, other):
        if type(other) == str:
            other = Monomial.as_var(other)
        return not (self == other)

    def __lt__(self, other):
        self_vars = set(self._power_dict.keys())
        if type(other) == str:
            other = Monomial.as_var(other)
        other_vars = set(other._power_dict.keys())
        # if we have a var that they don't we cannot be "smaller"
        if len(self_vars - other_vars) > 0:
            return False
        # check that we do not exceed and are smaller at least once
        at_least_once_less = False
        for var in self_vars:
            if self._power_dict[var] > other._power_dict[var]:
                return False
            elif self._power_dict[var] < other._power_dict[var]:
                at_least_once_less = True
        return at_least_once_less or len(other_vars - self_vars) > 0

    def __le__(self, other):
        self_vars = set(self._power_dict.keys())
        if type(other) == str:
            other = Monomial.as_var(other)
        other_vars = set(other._power_dict.keys())
        # if we have a var that they don't we cannot be "smaller"
        if len(self_vars - other_vars) > 0:
            return False
        # check that we do not exceed
        for var in self_vars:
            if self._power_dict[var] > other._power_dict[var]:
                return False
        return True

    def __gt__(self, other):
        self_vars = set(self._power_dict.keys())
        if type(other) == str:
            other = Monomial.as_var(other)
        other_vars = set(other._power_dict.keys())
        # if they have a var that they don't we cannot be "greater"
        if len(other_vars - self_vars) > 0:
            return False
        # check that we are not smaller and are greater at least once
        at_least_once_greater = False
        for var in other_vars:
            if self._power_dict[var] < other._power_dict[var]:
                return False
            elif self._power_dict[var] > other._power_dict[var]:
                at_least_once_greater = True
        return at_least_once_greater or len(self_vars - other_vars) > 0

    def __ge__(self, other):
        self_vars = set(self._power_dict.keys())
        if type(other) == str:
            other = Monomial.as_var(other)
        other_vars = set(other._power_dict.keys())
        # if they have a var that they don't we cannot be "greater"
        if len(other_vars - self_vars) > 0:
            return False
        # check that we are not smaller
        for var in other_vars:
            if self._power_dict[var] < other._power_dict[var]:
                return False
        return True

    def __hash__(self):
        return sum(hash(k) for k in self._power_dict.keys()) + sum(
            hash(v) for v in self._power_dict.values()
        )

    def __str__(self):
        if self._power_dict == {}:
            return "1"
        else:
            variables = sorted(self._power_dict.keys())
            return "*".join(
                [
                    str(var) + "^" + str(self._power_dict[var])
                    if self._power_dict[var] > 1
                    else str(var)
                    for var in variables
                ]
            )

    __repr__ = __str__

    def as_c_expression(self):
        if self._power_dict == {}:
            return "1"
        else:
            variables = sorted(self._power_dict.keys())
            return "*".join(
                [
                    "mod3pow(" + str(var) + "," + str(self._power_dict[var]) + ")"
                    if self._power_dict[var] > 1
                    else str(var)
                    for var in variables
                    if self._power_dict[var] != 0
                ]
            )

    # def as_sympy(self):
    #     # sympy empty product is 1, consistent with power_dict
    #     return sympy.prod([sympy.Symbol(var, integer=True) ** pow
    #                        for var, pow in self._power_dict.items()])
    # Fun fact: sympy doesn't recognize Symbol(var) and Symbol(var, integer=True) to be the same

    def as_numpy_str(self, variables) -> str:
        if len(self._power_dict) == 0:
            return "1"

        return (
            "("
            + "*".join(
                [
                    "1".format(variables.index(var), self._power_dict[var])
                    if self._power_dict[var] == 0
                    else "state[{0}]".format(variables.index(var))
                    if self._power_dict[var] == 1
                    else "(state[{0}]**{1})".format(
                        variables.index(var), self._power_dict[var]
                    )
                    for var in self._power_dict
                ]
            )
            + ")"
        )

    def _make_inner_mathml(self) -> Tag:
        if len(self._power_dict) == 0:
            return inner_mathml_constant(1)

        if len(self._power_dict) == 1:
            # we don't need an enclosing product when there is only a single variable^power
            var, power = list(self._power_dict.items())[0]
            return single_var_to_power_inner_mathml(var, power)

        apply_tag = Tag(name="apply", is_xml=True)
        product_tag = Tag(name="times", is_xml=True, can_be_empty_element=True)
        apply_tag.append(product_tag)

        for var, power in self._power_dict.items():
            apply_tag.append(single_var_to_power_inner_mathml(var, power))

        return apply_tag


####################################################################################################


@attrs(init=False, repr=False, str=False, eq=False)
class Mod3Poly(Expression):
    """a sparse polynomial class"""

    coeff_dict: Dict[Monomial, int] = attrib()

    def __init__(self, coeffs: Union[Dict, int]):
        if type(coeffs) == dict:
            self.coeff_dict = {
                monomial: coeffs[monomial]
                for monomial in coeffs
                if coeffs[monomial] != 0
            }
        elif is_integer(coeffs):
            self.coeff_dict = {Monomial.unit(): (coeffs % 3)}
        else:
            raise TypeError(
                "unsupported initialization type for '{}': '{}'".format(
                    self.__class__, type(coeffs)
                )
            )

    def rename_variables(self, name_dict: Dict[str, str]):
        return Mod3Poly(
            coeffs={
                monomial.rename_variables(name_dict): coeff
                for monomial, coeff in self.coeff_dict.items()
            }
        )

    @staticmethod
    def zero():
        return Mod3Poly({Monomial.unit(): 0})

    @staticmethod
    def one():
        return Mod3Poly({Monomial.unit(): 1})

    def as_polynomial(self):
        return self

    def __int__(self):
        self.__clear_zero_monomials()
        if len(self.coeff_dict) > 1 or (
            len(self.coeff_dict) == 1 and Monomial.unit() not in self.coeff_dict
        ):
            raise Exception("cannot cast non-constant polynomial to int")
        if Monomial.unit() in self.coeff_dict:
            return self.coeff_dict[Monomial.unit()]
        else:
            return 0

    def eval(self, variable_dict: Dict) -> Expression:
        """
        Evaluates the polynomial.

        Parameters
        ----------
        variable_dict
            variable_dict is expected to be a dict containing str:Expression or
            Monomial:Expression pairs. The latter are constrained to be of single-variable type.
        Returns
        -------
        Expression
        """
        if type(variable_dict) != dict:
            raise Exception("Mod3Poly.eval is not defined on this input")

        accumulator = Mod3Poly.zero()
        for monomial, coeff in self.coeff_dict.items():
            accumulator += coeff * monomial.eval(variable_dict)
        return accumulator

    def get_variable_set(self) -> Set[str]:
        """return a set containing all variables which occur in this polynomial"""
        var_set = set()
        for monomial in self.coeff_dict:
            var_set = var_set.union(monomial.get_variable_set())
        return var_set

    def __clear_zero_monomials(self):
        """purge unneeded data"""
        self.coeff_dict = {
            monomial: self.coeff_dict[monomial]
            for monomial in self.coeff_dict
            if self.coeff_dict[monomial] != 0
        }
        # assure at least one entry
        if len(self.coeff_dict) == 0:
            self.coeff_dict = {Monomial.unit(): 0}

    def is_constant(self):
        # possibly unnecessary
        self.__clear_zero_monomials()
        num_nonzero_monomial = len(self.coeff_dict)
        if num_nonzero_monomial > 1:
            return False
        elif num_nonzero_monomial == 0:
            return True
        else:
            # only one entry
            return Monomial.unit() in self.coeff_dict

    def __eq__(self, other):
        if isinstance(other, int):
            return self.is_constant() and int(self) == other
        elif not isinstance(other, Expression):
            return False

        if not isinstance(other, Mod3Poly):
            other_as_poly: Mod3Poly = other.as_poly()
        else:
            other_as_poly = other

        for term in self.coeff_dict:
            if (
                term not in other_as_poly.coeff_dict
                or (self.coeff_dict[term] - other_as_poly.coeff_dict[term]) % 3 != 0
            ):
                return False

        for term in other_as_poly.coeff_dict:
            if (
                term not in self.coeff_dict
                or (self.coeff_dict[term] - other_as_poly.coeff_dict[term]) % 3 != 0
            ):
                return False

        return True

    def __getitem__(self, index):
        if index in self.coeff_dict:
            return self.coeff_dict[index]
        else:
            return 0

    def __setitem__(self, index, value):
        self.coeff_dict[index] = value

    def __add__(self, other):
        if is_integer(other):
            self_copy = Mod3Poly(self.coeff_dict)
            self_copy[Monomial.unit()] = (self_copy[Monomial.unit()] + other) % 3
            return self_copy
        elif isinstance(other, Monomial):
            self_copy = Mod3Poly(self.coeff_dict)
            self_copy[other] += 1
            return self_copy
        elif isinstance(other, Mod3Poly):
            self_copy = Mod3Poly(self.coeff_dict)
            for key in other.coeff_dict.keys():
                if key in self_copy.coeff_dict.keys():
                    self_copy[key] = (self_copy[key] + other[key]) % 3
                else:
                    self_copy[key] = other[key]
            return self_copy
        elif isinstance(other, Expression):
            return BinaryOperation("PLUS", self, other)
        else:
            raise TypeError(
                "unsupported operand type(s) for +: '{}' and '{}'".format(
                    self.__class__, type(other)
                )
            )

    __radd__ = __add__

    def __sub__(self, other):
        if is_integer(other):
            self_copy = Mod3Poly(self.coeff_dict)
            self_copy[Monomial.unit()] = (self_copy[Monomial.unit()] - other) % 3
            return self_copy
        elif isinstance(other, Mod3Poly) or isinstance(other, Monomial):
            self_copy = Mod3Poly(self.coeff_dict)
            if isinstance(other, Monomial):
                other = other.as_poly()
            for key in other.coeff_dict.keys():
                if key in self_copy.coeff_dict.keys():
                    self_copy[key] = (self_copy[key] - other[key]) % 3
                else:
                    self_copy[key] = other[key]
            return self_copy
        else:
            raise TypeError(
                "unsupported operand type(s) for +: '{}' and '{}'".format(
                    self.__class__, type(other)
                )
            )

    def __rsub__(self, other):
        return other + ((-1) * self)

    def __mul__(self, other):
        if is_integer(other):
            return Mod3Poly(
                {key: (self.coeff_dict[key] * other) % 3 for key in self.coeff_dict}
            )
        elif isinstance(other, Monomial):
            return Mod3Poly(
                {
                    (other * monomial): self.coeff_dict[monomial]
                    for monomial in self.coeff_dict
                }
            )
        elif isinstance(other, Mod3Poly):
            accumulator = Mod3Poly.zero()
            for self_mono, other_mono in product(
                self.coeff_dict.keys(), other.coeff_dict.keys()
            ):
                monomial_prod = self_mono * other_mono
                accumulator[monomial_prod] = (
                    accumulator[monomial_prod] + self[self_mono] * other[other_mono]
                ) % 3
            return accumulator
        else:
            return BinaryOperation("TIMES", self, other)

    __rmul__ = __mul__

    def __pow__(self, power, **kwargs):
        if type(power) == Mod3Poly and power.is_constant():
            power = power[Monomial.unit()]
        assert is_integer(power)
        if power == 0:
            return Monomial.unit().as_poly()
        elif power == 1:
            return self
        elif power == 2:
            return self * self
        # Now handle higher powers; probably not going to happen too much for this application

        # (int) half power root
        int_root = self ** (power // 2)
        if power % 2 == 0:
            return int_root * int_root
        else:
            return int_root * int_root * self

    def __str__(self):
        accumulator = ""
        for monomial in sorted(self.coeff_dict.keys()):
            if monomial == Monomial.unit():
                if self[monomial] != 0:
                    accumulator += str(self[monomial])
            else:
                if len(accumulator) > 0 and self[monomial] != 0:
                    accumulator += "+"
                if self[monomial] == 1:
                    accumulator += str(monomial)
                elif self[monomial] == 2:
                    accumulator += "2*"
                    accumulator += str(monomial)
        if len(accumulator) > 0:
            return accumulator
        else:
            return "0"

    __repr__ = __str__

    def as_c_expression(self):
        accumulator = ""
        for monomial in sorted(self.coeff_dict.keys()):
            if monomial == Monomial.unit():
                if self[monomial] != 0:
                    accumulator += str(self[monomial])
            else:
                if len(accumulator) > 0 and self[monomial] != 0:
                    accumulator += "+"
                if self[monomial] == 1:
                    accumulator += monomial.as_c_expression()
                elif self[monomial] == 2:
                    accumulator += "2*"
                    accumulator += monomial.as_c_expression()
        if len(accumulator) > 0:
            return accumulator
        else:
            return "0"

    # def as_sympy(self):
    #     return sum([coeff * expr.as_sympy() for expr, coeff in self.coeff_dict.items()])

    def as_numpy_str(self, variables) -> str:
        return (
            "("
            + "+".join(
                [
                    "({0}*({1}))".format(coeff, expr.as_numpy_str(variables))
                    for expr, coeff in self.coeff_dict.items()
                ]
            )
            + ")"
        )

    def _make_inner_mathml(self) -> Tag:

        num_terms = len(self.coeff_dict)

        if num_terms == 0:
            return inner_mathml_constant(0)
        elif num_terms == 1:
            monomial, coefficient = list(self.coeff_dict.items())[0]
            if monomial.is_constant():
                return inner_mathml_constant(coefficient)
            apply_tag = Tag(name="apply", is_xml=True)
            apply_tag.append(Tag(name="times", is_xml=True, can_be_empty_element=True))
            apply_tag.append(inner_mathml_constant(coefficient))
            # noinspection PyProtectedMember
            apply_tag.append(monomial._make_inner_mathml())
            return apply_tag
        else:
            top_apply_tag = Tag(name="apply", is_xml=True)
            top_apply_tag.append(
                Tag(name="plus", is_xml=True, can_be_empty_element=True)
            )
            for monomial, coefficient in self.coeff_dict.items():
                if coefficient == 0:
                    top_apply_tag.append(inner_mathml_constant(0))
                elif coefficient == 1:
                    # noinspection PyProtectedMember
                    top_apply_tag.append(monomial._make_inner_mathml())
                elif monomial.is_constant():
                    top_apply_tag.append(inner_mathml_constant(coefficient))
                else:
                    inner_apply_tag = Tag(name="apply", is_xml=True)
                    inner_apply_tag.append(
                        Tag(name="times", is_xml=True, can_be_empty_element=True)
                    )
                    inner_apply_tag.append(inner_mathml_constant(coefficient))
                    # noinspection PyProtectedMember
                    inner_apply_tag.append(monomial._make_inner_mathml())
                    top_apply_tag.append(inner_apply_tag)
            return top_apply_tag



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read().splitlines()

setup(
    name="steady_cell_phenotype",
    version="0.3.1",
    author="A. C. Knapp",
    author_email="adam.knapp@medicine.ufl.edu",
    description="A tool for the computation of steady states and exploration of dynamics "
    "in ternary intracellular biological networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/knappa/steadycellphenotype",
    project_urls={
        "Bug Tracker": "https://github.com/knappa/steadycellphenotype/issues",
    },
    packages=find_packages(),
    include_package_data=True,
    scripts=["steady_cell_phenotype/scp_converter.py", "start_scp.sh"],
    zip_safe=False,
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.7",
)
